# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import glob
import logging
import os
import re
import shutil
import subprocess
import zipfile
from concurrent.futures import ThreadPoolExecutor, as_completed
from multiprocessing import cpu_count
from pathlib import Path, PosixPath
from subprocess import Popen
from typing import List

import pendulum
from airflow.exceptions import AirflowException
from airflow.hooks.base import BaseHook
from airflow.models.taskinstance import TaskInstance
from airflow.models.variable import Variable
from google.cloud import storage
from google.cloud.bigquery import SourceFormat
from google.cloud.storage import Blob
from mag_archiver.mag import MagArchiverClient, MagDateType, MagRelease, MagState
from natsort import natsorted

from academic_observatory_workflows.config import schema_folder
from observatory.platform.utils.airflow_utils import AirflowConns
from observatory.platform.utils.airflow_utils import (
    AirflowVars,
    check_connections,
    check_variables,
)
from observatory.platform.utils.config_utils import find_schema
from observatory.platform.utils.gc_utils import (
    azure_to_google_cloud_storage_transfer,
    bigquery_sharded_table_id,
    bigquery_table_exists,
    create_bigquery_dataset,
    download_blobs_from_cloud_storage,
    load_bigquery_table,
    table_name_from_blob,
    upload_files_to_cloud_storage,
)
from observatory.platform.utils.proc_utils import wait_for_process
from observatory.platform.utils.workflow_utils import (
    SubFolder,
    workflow_path,
)


def pull_releases(ti: TaskInstance) -> List[MagRelease]:
    """Pull a list of MagRelease instances with xcom.

    :param ti: the Apache Airflow task instance.
    :return: the list of MagRelease instances.
    """

    return ti.xcom_pull(
        key=MagTelescope.RELEASES_TOPIC_NAME, task_ids=MagTelescope.TASK_ID_LIST, include_prior_dates=False
    )


def list_mag_release_files(release_path: str) -> List[PosixPath]:
    """List the MAG release file paths in a particular folder. Excludes the samples directory.
    :param release_path: the path to the MAG release.
    :return: a list of PosixPath files.
    """

    release_folder = os.path.basename(os.path.abspath(release_path))
    include_regex = fr"^.*/{release_folder}(/advanced|/mag|/nlp)?/\w+.txt(.[0-9]+)?$"

    types = ["*.txt", "*.txt.[0-9]"]
    files = []
    for file_type in types:
        paths = list(Path(release_path).rglob(file_type))
        for path in paths:
            path_string = str(path.resolve())
            if re.match(include_regex, path_string) is not None:
                files.append(path)
    files = natsorted(files, key=lambda x: str(x))
    return files


def transform_mag_file(input_file_path: str, output_file_path: str) -> bool:
    r"""Transform MAG file, removing the \x0 and \r characters. \r is the ^M windows character.

    :param input_file_path: the path of the file to transform.
    :param output_file_path: where to save the transformed file.
    :return: whether the transformation was successful or not.
    """

    # TODO: see if we can get rid of shell=True
    bash_command = fr"sed 's/\r//g; s/\x0//g' {str(input_file_path)} > {output_file_path}"
    logging.info(f"transform_mag_file bash command: {bash_command}")
    proc: Popen = subprocess.Popen(bash_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    output, error = wait_for_process(proc)
    logging.debug(output)
    success = proc.returncode == 0

    if success:
        logging.info(f"transform_mag_file success: {input_file_path}")
    else:
        logging.error(f"transform_mag_file error: {input_file_path}")
        logging.error(error)

    return success


def transform_mag_release(input_release_path: str, output_release_path: str, max_workers: int = cpu_count()) -> bool:
    """Transform a MAG release into a form that can be loaded into BigQuery.

    :param input_release_path: the path to the folder containing the files for the MAG release.
    :param output_release_path: the path where the transformed files will be saved.
    :param max_workers: the number of processes to use when transforming files (one process per file).
    :return: whether the transformation was successful or not.
    """

    # Transform each file in parallel
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        # Create tasks
        futures = []
        futures_msgs = {}

        paths = list_mag_release_files(input_release_path)
        for path in paths:
            # Make path to save file
            os.makedirs(output_release_path, exist_ok=True)
            output_path = os.path.join(output_release_path, path.name)
            msg = f"input_file_path={path}, output_file_path={output_path}"
            logging.info(f"transform_mag_release: {msg}")
            future = executor.submit(transform_mag_file, path, output_path)
            futures.append(future)
            futures_msgs[future] = msg

        # Wait for completed tasks
        results = []
        for future in as_completed(futures):
            success = future.result()
            msg = futures_msgs[future]
            results.append(success)
            if success:
                logging.info(f"transform_mag_release success: {msg}")
            else:
                logging.error(f"transform_mag_release failed: {msg}")

    return all(results)


class MagTelescope:
    """A container for holding the constants and static functions for the Microsoft Academic Graph (MAG) telescope.

    Requires the following connections to be added to Airflow:
        mag_releases_table: the Azure account name (login) and sas token (password) for the MagReleases table in
        Azure.
        mag_snapshots_container: the Azure Storage Account name (login) and the sas token (password) for the
        Azure storage blob container that contains the MAG releases.
    """

    DAG_ID = "mag"
    DATASET_ID = "mag"
    QUEUE = "remote_queue"
    DESCRIPTION = (
        "The Microsoft Academic Graph (MAG) dataset: https://www.microsoft.com/en-us/research/project/"
        "microsoft-academic-graph/"
    )
    RELEASES_TOPIC_NAME = "releases"
    MAX_PROCESSES = cpu_count()
    MAX_CONNECTIONS = cpu_count()
    RETRIES = 3

    TASK_ID_CHECK_DEPENDENCIES = "check_dependencies"
    TASK_ID_LIST = "list_releases"
    TASK_ID_TRANSFER = "transfer"
    TASK_ID_DOWNLOAD = "download"
    TASK_ID_TRANSFORM = "transform"
    TASK_ID_UPLOAD_TRANSFORMED = "upload_transformed"
    TASK_ID_BQ_LOAD = "bq_load"
    TASK_ID_CLEANUP = "cleanup"

    @staticmethod
    def check_dependencies(**kwargs):
        """Check that all variables and connections exist that are required to run the DAG.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        vars_valid = check_variables(
            AirflowVars.DATA_PATH,
            AirflowVars.PROJECT_ID,
            AirflowVars.DATA_LOCATION,
            AirflowVars.DOWNLOAD_BUCKET,
            AirflowVars.TRANSFORM_BUCKET,
        )
        conns_valid = check_connections(AirflowConns.MAG_RELEASES_TABLE, AirflowConns.MAG_SNAPSHOTS_CONTAINER)

        if not vars_valid or not conns_valid:
            raise AirflowException("Required variables or connections are missing")

    @staticmethod
    def list_releases(**kwargs):
        """Task to list all MAG releases for a given month.

        Requires the following connection to be added to Airflow:
            mag_releases_table: the Azure account name (login) and the sas token (password) for the MagReleases table in
            Azure.

        Pushes the following xcom:
            a list of MagRelease instances.

        :param kwargs: the context passed from the BranchPythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: the identifier of the task to execute next.
        """

        connection = BaseHook.get_connection("mag_releases_table")
        account_name = connection.login
        sas_token = connection.password
        execution_date = kwargs["execution_date"]
        next_execution_date = kwargs["next_execution_date"]
        project_id = Variable.get(AirflowVars.PROJECT_ID)

        client = MagArchiverClient(account_name=account_name, sas_token=sas_token)
        releases: List[MagRelease] = client.list_releases(
            start_date=execution_date, end_date=next_execution_date, state=MagState.done, date_type=MagDateType.done
        )

        # Check if we can skip any releases
        releases_out = []
        logging.info("Check if releases already exist:")
        for release in releases:
            table_id = bigquery_sharded_table_id(MagTelescope.DAG_ID, release.release_date)

            if bigquery_table_exists(project_id, MagTelescope.DATASET_ID, table_id):
                logging.info(
                    f"Skipping as table exists for MAG {release.release_date} release: "
                    f"{project_id}.{MagTelescope.DATASET_ID}.{table_id}"
                )
            else:
                logging.info(f"Table doesn't exist yet, processing MAG {release.release_date} release in this workflow")
                releases_out.append(release)

        continue_dag = len(releases_out)
        if continue_dag:
            # Push messages
            ti: TaskInstance = kwargs["ti"]
            ti.xcom_push(MagTelescope.RELEASES_TOPIC_NAME, releases_out, execution_date)
        return continue_dag

    @staticmethod
    def transfer(**kwargs):
        """Task to transfer a MAG release from Azure to Google Cloud Storage.

        Requires the following connection to be added to Airflow:
            mag_snapshots_container: the Azure Storage Account name (login) and the sas token (password) for the
            Azure storage blob container that contains the MAG releases.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Get MAG releases
        ti: TaskInstance = kwargs["ti"]
        releases = pull_releases(ti)

        # Get variables
        gcp_project_id = Variable.get(AirflowVars.PROJECT_ID)
        gcp_bucket_name = Variable.get(AirflowVars.DOWNLOAD_BUCKET)

        # Get Azure connection information
        connection = BaseHook.get_connection("mag_snapshots_container")
        azure_account_name = connection.login
        azure_sas_token = connection.password

        # Download and extract each release posted this month
        azure_container = None
        row_keys = []
        include_prefixes = []

        for release in releases:
            azure_container = release.release_container
            row_keys.append(release.row_key)
            include_prefixes.append(release.release_path)

        if len(row_keys) > 0:
            description = "Transfer MAG Releases: " + ", ".join(row_keys)
            logging.info(description)
            success = azure_to_google_cloud_storage_transfer(
                azure_account_name,
                azure_sas_token,
                azure_container,
                include_prefixes,
                gcp_project_id,
                gcp_bucket_name,
                description,
            )
            releases_str = ", ".join(row_keys)
            if success:
                logging.info(f"Success transferring MAG releases: {releases_str}")
            else:
                logging.error(f"Error transferring MAG release: {releases_str}")
                exit(os.EX_DATAERR)
        else:
            logging.warning("No release to transfer")

    @staticmethod
    def download(**kwargs):
        """Downloads the MAG release from Google Cloud Storage.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Get variables
        bucket_name = Variable.get(AirflowVars.DOWNLOAD_BUCKET)

        # Get MAG releases
        ti: TaskInstance = kwargs["ti"]
        releases = pull_releases(ti)

        # Download each release to the extracted folder path (since they are already extracted)
        extracted_path = workflow_path(SubFolder.extracted, MagTelescope.DAG_ID)
        for release in releases:
            logging.info(f"Downloading release: {release}")
            destination_path = os.path.join(extracted_path, release.source_container)
            success = download_blobs_from_cloud_storage(
                bucket_name,
                release.release_path,
                destination_path,
                max_processes=MagTelescope.MAX_PROCESSES,
                max_connections=MagTelescope.MAX_CONNECTIONS,
                retries=MagTelescope.RETRIES,
            )

            if success:
                logging.info(f"Success downloading MAG release: {release}")
            else:
                logging.error(f"Error downloading MAG release: {release}")
                exit(os.EX_DATAERR)

    @staticmethod
    def transform(**kwargs):
        """Transforms the MAG release into a form that can be uploaded to BigQuery.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Get MAG releases
        ti: TaskInstance = kwargs["ti"]
        releases = pull_releases(ti)

        # For each release and folder to include, transform the files with sed and save into the transformed directory
        for release in releases:
            logging.info(f"Transforming MAG release: {release}")
            release_extracted_path = os.path.join(
                workflow_path(SubFolder.extracted, MagTelescope.DAG_ID), release.source_container
            )
            release_transformed_path = os.path.join(
                workflow_path(SubFolder.transformed, MagTelescope.DAG_ID), release.source_container
            )
            success = transform_mag_release(
                release_extracted_path, release_transformed_path, max_workers=MagTelescope.MAX_PROCESSES
            )

            if success:
                logging.info(f"Success transforming MAG release: {release}")
            else:
                logging.error(f"Error transforming MAG release: {release}")
                exit(os.EX_DATAERR)

    @staticmethod
    def upload_transformed(**kwargs):
        """Uploads the transformed MAG release files to Google Cloud Storage for loading into BigQuery.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Get variables
        bucket_name = Variable.get(AirflowVars.TRANSFORM_BUCKET)

        # Get MAG releases
        ti: TaskInstance = kwargs["ti"]
        releases = pull_releases(ti)

        # Upload files to cloud storage
        for release in releases:
            logging.info(f"Uploading MAG release to cloud storage: {release}")
            release_transformed_path = os.path.join(
                workflow_path(SubFolder.transformed, MagTelescope.DAG_ID), release.source_container
            )
            posix_paths = list_mag_release_files(release_transformed_path)
            paths = [str(path) for path in posix_paths]
            blob_names = [
                f"telescopes/{MagTelescope.DAG_ID}/{release.source_container}/{path.name}" for path in posix_paths
            ]
            success = upload_files_to_cloud_storage(
                bucket_name,
                blob_names,
                paths,
                max_processes=MagTelescope.MAX_PROCESSES,
                max_connections=MagTelescope.MAX_CONNECTIONS,
                retries=MagTelescope.RETRIES,
            )
            if success:
                logging.info(f"Success uploading MAG release to cloud storage: {release}")
            else:
                logging.error(f"Error uploading MAG release to cloud storage: {release}")
                exit(os.EX_DATAERR)

    @staticmethod
    def bq_load(**kwargs):
        """Loads a MAG release into BigQuery.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Get MAG releases
        ti: TaskInstance = kwargs["ti"]
        releases = pull_releases(ti)

        # Get config variables
        project_id = Variable.get(AirflowVars.PROJECT_ID)
        data_location = Variable.get(AirflowVars.DATA_LOCATION)
        bucket_name = Variable.get(AirflowVars.TRANSFORM_BUCKET)

        # For each release, load into BigQuery
        for release in releases:
            release_path = f"telescopes/{MagTelescope.DAG_ID}/{release.source_container}"
            success = db_load_mag_release(project_id, bucket_name, data_location, release_path, release.release_date)

            if success:
                logging.info(f"Success loading MAG release: {release}")
            else:
                logging.error(f"Error loading MAG release: {release}")
                exit(os.EX_DATAERR)

    @staticmethod
    def cleanup(**kwargs):
        """Delete files of downloaded, extracted and transformed releases.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Pull releases
        ti: TaskInstance = kwargs["ti"]
        releases = pull_releases(ti)

        for release in releases:
            # Remove all extracted files
            release_extracted_path = os.path.join(
                workflow_path(SubFolder.extracted, MagTelescope.DAG_ID), release.source_container
            )
            try:
                shutil.rmtree(release_extracted_path)
            except FileNotFoundError as e:
                logging.warning(f"No such file or directory {release_extracted_path}: {e}")

            # Remove all transformed files
            release_transformed_path = os.path.join(
                workflow_path(SubFolder.transformed, MagTelescope.DAG_ID), release.source_container
            )
            try:
                shutil.rmtree(release_transformed_path)
            except FileNotFoundError as e:
                logging.warning(f"No such file or directory {release_transformed_path}: {e}")


def db_load_mag_release(
    project_id: str,
    bucket_name: str,
    data_location: str,
    release_path: str,
    release_date: pendulum.DateTime,
    dataset_id: str = MagTelescope.DAG_ID,
) -> bool:
    """Load a MAG release into BigQuery.

    :param project_id: the Google Cloud project id.
    :param bucket_name: the Google Cloud bucket name where the transformed files are stored.
    :param data_location: the location where the BigQuery dataset will be created.
    :param release_path: the path on the Google Cloud storage bucket where the particular MAG release is located.
    :param release_date: the release date of the MAG release.
    :param dataset_id: the identifier of the dataset.
    :return: whether the MAG release was loaded into BigQuery successfully.
    """

    settings = {
        "Authors": {"quote": "", "allow_quoted_newlines": True},
        "FieldsOfStudy": {"quote": "", "allow_quoted_newlines": False},
        "PaperAuthorAffiliations": {"quote": "", "allow_quoted_newlines": False},
        "PaperCitationContexts": {"quote": "", "allow_quoted_newlines": True},
        "PaperExtendedAttributes": {"quote": "", "allow_quoted_newlines": False},
        "Papers": {"quote": "", "allow_quoted_newlines": True},
    }

    # Create dataset
    create_bigquery_dataset(project_id, dataset_id, data_location, MagTelescope.DESCRIPTION)

    # Get bucket
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)

    # List release blobs
    blobs: List[Blob] = list(bucket.list_blobs(prefix=release_path))
    max_workers = len(blobs)

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        # Create tasks
        futures = []
        futures_msgs = {}
        analysis_schema_path = schema_folder()
        prefix = "Mag"
        file_extension = ".txt"

        # De-duplicate blobs, i.e. for tables where there are more than one file:
        # e.g. PaperAbstractsInvertedIndex.txt.1 and PaperAbstractsInvertedIndex.txt.2 become
        # PaperAbstractsInvertedIndex.txt.* so that both are loaded into the same table.
        blob_names = set()
        for blob in blobs:
            blob_name = blob.name
            if not blob_name.endswith(file_extension):
                blob_name_sans_index = re.match(r"^.+?(?=([0-9]+)?$)", blob_name).group(0)
                blob_name_with_wildcard = f"{blob_name_sans_index}*"
                blob_names.add(blob_name_with_wildcard)
            else:
                blob_names.add(blob_name)

        for blob_name in blob_names:
            # Make table name and id
            table_name = table_name_from_blob(blob_name, file_extension)
            table_id = bigquery_sharded_table_id(table_name, release_date)

            # Get schema for table
            schema_file_path = find_schema(analysis_schema_path, table_name, release_date, prefix=prefix)
            if schema_file_path is None:
                logging.error(
                    f"No schema found with search parameters: analysis_schema_path={analysis_schema_path}, "
                    f"table_name={table_name}, release_date={release_date}, prefix={prefix}"
                )
                exit(os.EX_CONFIG)

            uri = f"gs://{bucket_name}/{blob_name}"
            msg = f"uri={uri}, table_id={table_id}, schema_file_path={schema_file_path}"
            logging.info(f"db_load_mag_release: {msg}")

            if table_name in settings:
                csv_quote_character = settings[table_name]["quote"]
                csv_allow_quoted_newlines = settings[table_name]["allow_quoted_newlines"]
            else:
                csv_quote_character = '"'
                csv_allow_quoted_newlines = False

            future = executor.submit(
                load_bigquery_table,
                uri,
                dataset_id,
                data_location,
                table_id,
                schema_file_path,
                SourceFormat.CSV,
                csv_field_delimiter="\t",
                csv_quote_character=csv_quote_character,
                csv_allow_quoted_newlines=csv_allow_quoted_newlines,
            )
            futures_msgs[future] = msg
            futures.append(future)

        # Wait for completed tasks
        results = []
        for future in as_completed(futures):
            success = future.result()
            msg = futures_msgs[future]
            results.append(success)
            if success:
                logging.info(f"db_load_mag_release success: {msg}")
            else:
                logging.error(f"db_load_mag_release failed: {msg}")

    return all(results)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs

import logging
import os
import pathlib
import re
import shutil
import subprocess
from typing import List

import pendulum
import xmltodict
from airflow.exceptions import AirflowException
from airflow.models import Variable
from airflow.models.taskinstance import TaskInstance
from google.cloud.bigquery import SourceFormat

from academic_observatory_workflows.config import schema_folder
from observatory.platform.utils.airflow_utils import AirflowVars, check_variables
from observatory.platform.utils.config_utils import find_schema
from observatory.platform.utils.data_utils import get_file
from observatory.platform.utils.gc_utils import (
    bigquery_sharded_table_id,
    bigquery_table_exists,
    create_bigquery_dataset,
    load_bigquery_table,
    upload_file_to_cloud_storage,
)
from observatory.platform.utils.proc_utils import wait_for_process
from observatory.platform.utils.url_utils import retry_session
from observatory.platform.utils.workflow_utils import (
    SubFolder,
    workflow_path,
)


def pull_releases(ti: TaskInstance) -> List:
    """Pull a list of Unpaywall release instances with xcom.

    :param ti: the Apache Airflow task instance.
    :return: the list of Unpaywall release instances.
    """

    return ti.xcom_pull(
        key=UnpaywallTelescope.RELEASES_TOPIC_NAME, task_ids=UnpaywallTelescope.TASK_ID_LIST, include_prior_dates=False
    )


def list_releases(start_date: pendulum.DateTime, end_date: pendulum.DateTime) -> List["UnpaywallRelease"]:
    """Parses xml string retrieved from GET request to create list of urls for
    different releases.

    :param start_date:
    :param end_date:
    :return: a list of UnpaywallRelease instances.
    """
    releases_list = []

    # Request releases page
    response = retry_session().get(UnpaywallTelescope.TELESCOPE_URL)

    if response is not None and response.status_code == 200:
        # Parse releases
        items = xmltodict.parse(response.text)["ListBucketResult"]["Contents"]
        for item in items:
            # Get filename and parse dates
            file_name = item["Key"]
            last_modified = pendulum.parse(item["LastModified"])
            release_date = UnpaywallRelease.parse_release_date(file_name)

            # Only include release if last modified date is within start and end date.
            # Last modified date is used rather than release date because if release date is used then releases will
            # be missed.
            if start_date <= last_modified < end_date:
                release = UnpaywallRelease(file_name, last_modified, release_date)
                releases_list.append(release)
    else:
        raise ConnectionError(f"Error requesting url: {UnpaywallTelescope.TELESCOPE_URL}")

    return releases_list


def download_release(release: "UnpaywallRelease") -> str:
    """Downloads release from url.

    :param release: Instance of UnpaywallRelease class
    """
    filename = release.get_filepath_download()
    download_dir = os.path.dirname(filename)

    # Download
    logging.info(f"Downloading file: {filename}, url: {release.url}")
    file_path, updated = get_file(fname=filename, origin=release.url, cache_dir=download_dir)

    if file_path:
        logging.info(f"Success downloading release: {filename}")
    else:
        logging.error(f"Error downloading release: {filename}")
        exit(os.EX_DATAERR)

    return file_path


def extract_release(release: "UnpaywallRelease") -> str:
    """Decompresses release.

    :param release: Instance of UnpaywallRelease class
    """
    logging.info(f"Extracting file: {release.filepath_download}")

    cmd = f"gunzip -c {release.filepath_download} > {release.filepath_extract}"
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, executable="/bin/bash")
    stdout, stderr = wait_for_process(p)

    if stdout:
        logging.info(stdout)

    if stderr:
        raise AirflowException(f"bash command failed for {release.url}: {stderr}")

    logging.info(f"File extracted to: {release.filepath_extract}")

    return release.filepath_extract


def transform_release(release: "UnpaywallRelease") -> str:
    """Transforms release by replacing a specific '-' with '_'.

    :param release: Instance of UnpaywallRelease class
    """
    cmd = (
        f"sed 's/authenticated-orcid/authenticated_orcid/g' {release.filepath_extract} > "
        f"{release.filepath_transform}"
    )

    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, executable="/bin/bash")
    stdout, stderr = wait_for_process(p)

    if stdout:
        logging.info(stdout)

    if stderr:
        raise AirflowException(f"bash command failed for {release.url}: {stderr}")

    logging.info(f"Success transforming release: {release.url}")

    return release.filepath_transform


class UnpaywallRelease:
    def __init__(self, file_name: str, last_modified: pendulum.DateTime, release_date: pendulum.DateTime):
        self.file_name = file_name
        self.last_modified = last_modified
        self.release_date = release_date
        self.filepath_download = self.get_filepath_download()
        self.filepath_extract = self.get_filepath_extract()
        self.filepath_transform = self.get_filepath_transform()

    @property
    def url(self):
        return f"{UnpaywallTelescope.TELESCOPE_URL}{self.file_name}"

    @staticmethod
    def parse_release_date(file_name: str) -> pendulum.DateTime:
        """Parses a release date from a file name.

        :return: date.
        """

        date = re.search(r"\d{4}-\d{2}-\d{2}", file_name).group()

        return pendulum.parse(date)

    def get_filepath_download(self) -> str:
        """Gives path to downloaded release.

        :return: absolute file path
        """

        date_str = self.release_date.strftime("%Y_%m_%d")
        compressed_file_name = f"{UnpaywallTelescope.DAG_ID}_{date_str}.jsonl.gz"
        download_dir = workflow_path(SubFolder.downloaded, UnpaywallTelescope.DAG_ID)
        path = os.path.join(download_dir, compressed_file_name)

        return path

    def get_filepath_extract(self) -> str:
        """Gives path to extracted and decompressed release.

        :return: absolute file path
        """

        date_str = self.release_date.strftime("%Y_%m_%d")
        decompressed_file_name = f"{UnpaywallTelescope.DAG_ID}_{date_str}.jsonl"
        extract_dir = workflow_path(SubFolder.extracted, UnpaywallTelescope.DAG_ID)
        path = os.path.join(extract_dir, decompressed_file_name)

        return path

    def get_filepath_transform(self) -> str:
        """Gives path to transformed release.

        :return: absolute file path
        """

        date_str = self.release_date.strftime("%Y_%m_%d")
        decompressed_file_name = f"{UnpaywallTelescope.DAG_ID}_{date_str}.jsonl"
        transform_dir = workflow_path(SubFolder.transformed, UnpaywallTelescope.DAG_ID)
        path = os.path.join(transform_dir, decompressed_file_name)

        return path


class UnpaywallTelescope:
    """A container for holding the constants and static functions for the Unpaywall telescope."""

    DAG_ID = "unpaywall"
    DATASET_ID = "our_research"
    QUEUE = "remote_queue"
    DESCRIPTION = "The Unpaywall database: https://unpaywall.org/"
    RELEASES_TOPIC_NAME = "releases"
    TELESCOPE_URL = "https://unpaywall-data-snapshots.s3-us-west-2.amazonaws.com/"
    RETRIES = 3

    TASK_ID_CHECK_DEPENDENCIES = "check_dependencies"
    TASK_ID_LIST = "list_releases"
    TASK_ID_STOP = "stop_dag"
    TASK_ID_DOWNLOAD = "download"
    TASK_ID_UPLOAD_DOWNLOADED = "upload_downloaded"
    TASK_ID_EXTRACT = "extract"
    TASK_ID_TRANSFORM = "transform"
    TASK_ID_UPLOAD_TRANSFORMED = "upload_transformed"
    TASK_ID_BQ_LOAD = "bq_load"
    TASK_ID_CLEANUP = "cleanup"

    @staticmethod
    def check_dependencies(**kwargs):
        """Check that all variables exist that are required to run the DAG.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        vars_valid = check_variables(
            AirflowVars.DATA_PATH,
            AirflowVars.PROJECT_ID,
            AirflowVars.DATA_LOCATION,
            AirflowVars.DOWNLOAD_BUCKET,
            AirflowVars.TRANSFORM_BUCKET,
        )
        if not vars_valid:
            raise AirflowException("Required variables are missing")

    @staticmethod
    def list_releases(**kwargs):
        """Based on a list of all releases, checks which ones were released between this and the next execution date
        of the DAG. If the release falls within the time period mentioned above, checks if a bigquery table doesn't
        exist yet for the release. A list of releases that passed both checks is passed to the next tasks. If the list
        is empty the workflow will stop.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Get variables
        project_id = Variable.get(AirflowVars.PROJECT_ID)

        # List releases between a start and end date
        execution_date = kwargs["execution_date"]
        next_execution_date = kwargs["next_execution_date"]
        releases_list = list_releases(execution_date, next_execution_date)
        logging.info(f"Releases between {execution_date} and {next_execution_date}:\n{releases_list}\n")

        # Check if the BigQuery table exists for each release to see if the workflow needs to process
        releases_list_out = []
        for release in releases_list:
            table_id = bigquery_sharded_table_id(UnpaywallTelescope.DAG_ID, release.release_date)

            if bigquery_table_exists(project_id, UnpaywallTelescope.DATASET_ID, table_id):
                logging.info(
                    f"Skipping as table exists for {release.url}: "
                    f"{project_id}.{UnpaywallTelescope.DATASET_ID}.{table_id}"
                )
            else:
                logging.info(f"Table doesn't exist yet, processing {release.url} in this workflow")
                releases_list_out.append(release)

        # If releases_list_out contains items then the DAG will continue (return True) otherwise it will
        # stop (return False)
        continue_dag = len(releases_list_out)
        if continue_dag:
            ti: TaskInstance = kwargs["ti"]
            ti.xcom_push(UnpaywallTelescope.RELEASES_TOPIC_NAME, releases_list_out)
        return continue_dag

    @staticmethod
    def download(**kwargs):
        """Download release to file.
        If develop environment, copy debug file from this repository to the right location. Else download from url.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Pull messages
        ti: TaskInstance = kwargs["ti"]
        releases_list = pull_releases(ti)

        # Download each release
        for release in releases_list:
            download_release(release)

    @staticmethod
    def upload_downloaded(**kwargs):
        """Upload downloaded release to a google cloud storage bucket.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Pull messages
        ti: TaskInstance = kwargs["ti"]
        releases_list = pull_releases(ti)

        # Get variables
        bucket_name = Variable.get(AirflowVars.DOWNLOAD_BUCKET)

        # Upload each release
        for release in releases_list:
            blob_name = f"telescopes/unpaywall/{os.path.basename(release.filepath_download)}"
            upload_file_to_cloud_storage(bucket_name, blob_name, file_path=release.filepath_download)

    @staticmethod
    def extract(**kwargs):
        """Unzip release to new file.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Pull messages
        ti: TaskInstance = kwargs["ti"]
        releases_list = pull_releases(ti)

        # Extract each release
        for release in releases_list:
            extract_release(release)

    @staticmethod
    def transform(**kwargs):
        """Transform release with sed command and save to new file.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Pull messages
        ti: TaskInstance = kwargs["ti"]
        releases_list = pull_releases(ti)

        # Transform each release
        for release in releases_list:
            transform_release(release)

    @staticmethod
    def upload_transformed(**kwargs):
        """Upload transformed release to a google cloud storage bucket.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Pull messages
        ti: TaskInstance = kwargs["ti"]
        releases_list = pull_releases(ti)

        # Get variables
        bucket_name = Variable.get(AirflowVars.TRANSFORM_BUCKET)

        # Upload each release
        for release in releases_list:
            blob_name = f"telescopes/unpaywall/{os.path.basename(release.filepath_transform)}"
            upload_file_to_cloud_storage(bucket_name, blob_name, file_path=release.filepath_transform)

    @staticmethod
    def load_to_bq(**kwargs):
        """Upload transformed release to a bigquery table.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Pull releases
        ti: TaskInstance = kwargs["ti"]
        releases_list: List[UnpaywallRelease] = pull_releases(ti)

        # Get variables
        project_id = Variable.get(AirflowVars.PROJECT_ID)
        bucket_name = Variable.get(AirflowVars.TRANSFORM_BUCKET)
        data_location = Variable.get(AirflowVars.DATA_LOCATION)

        # Create dataset
        create_bigquery_dataset(
            project_id, UnpaywallTelescope.DATASET_ID, data_location, UnpaywallTelescope.DESCRIPTION
        )

        for release in releases_list:
            # Get blob name
            blob_name = f"telescopes/unpaywall/{os.path.basename(release.filepath_transform)}"

            # Get release_date
            table_id = bigquery_sharded_table_id(UnpaywallTelescope.DAG_ID, release.release_date)

            # Select schema file based on release date
            analysis_schema_path = schema_folder()
            schema_file_path = find_schema(analysis_schema_path, UnpaywallTelescope.DAG_ID, release.release_date)
            if schema_file_path is None:
                logging.error(
                    f"No schema found with search parameters: analysis_schema_path={analysis_schema_path}, "
                    f"table_name={UnpaywallTelescope.DAG_ID}, release_date={release.release_date}"
                )
                exit(os.EX_CONFIG)

            # Load BigQuery table
            uri = f"gs://{bucket_name}/{blob_name}"
            logging.info(f"URI: {uri}")
            success = load_bigquery_table(
                uri,
                UnpaywallTelescope.DATASET_ID,
                data_location,
                table_id,
                schema_file_path,
                SourceFormat.NEWLINE_DELIMITED_JSON,
            )
            if not success:
                raise AirflowException("bq_load task: data failed to load data into BigQuery")

    @staticmethod
    def cleanup(**kwargs):
        """Delete files of downloaded, extracted and transformed release.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Pull releases
        ti: TaskInstance = kwargs["ti"]
        releases_list = pull_releases(ti)

        for release in releases_list:
            try:
                pathlib.Path(release.filepath_download).unlink()
            except FileNotFoundError as e:
                logging.warning(f"No such file or directory {release.filepath_download}: {e}")

            try:
                pathlib.Path(release.filepath_extract).unlink()
            except FileNotFoundError as e:
                logging.warning(f"No such file or directory {release.filepath_extract}: {e}")

            try:
                pathlib.Path(release.filepath_transform).unlink()
            except FileNotFoundError as e:
                logging.warning(f"No such file or directory {release.filepath_transform}: {e}")



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, James Diprose


from __future__ import annotations

import functools
import logging
import os
import shutil
import subprocess
from concurrent.futures import ThreadPoolExecutor, as_completed
from subprocess import Popen
from typing import Dict, List

import pendulum
import requests
from airflow.exceptions import AirflowException
from airflow.hooks.base import BaseHook
from bs4 import BeautifulSoup
from natsort import natsorted

from academic_observatory_workflows.config import schema_folder as default_schema_folder
from observatory.platform.utils.airflow_utils import AirflowConns, AirflowVars
from observatory.platform.utils.proc_utils import wait_for_process
from observatory.platform.utils.url_utils import retry_session
from observatory.platform.utils.workflow_utils import (
    blob_name,
    bq_load_shard,
    upload_files_from_list,
)
from observatory.platform.workflows.snapshot_telescope import (
    SnapshotRelease,
    SnapshotTelescope,
)


class CrossrefMetadataRelease(SnapshotRelease):
    def __init__(self, dag_id: str, release_date: pendulum.DateTime):
        """Create a CrossrefMetadataRelease instance.

        :param dag_id: the DAG id.
        :param release_date: the date of the release.
        """

        download_files_regex = ".*.json.tar.gz$"
        extract_files_regex = f".*.json$"
        transform_files_regex = f".*.jsonl$"
        super().__init__(dag_id, release_date, download_files_regex, extract_files_regex, transform_files_regex)

        self.url = CrossrefMetadataTelescope.TELESCOPE_URL.format(year=release_date.year, month=release_date.month)

    @property
    def download_path(self) -> str:
        """Get the path to the downloaded file.

        :return: the file path.
        """

        return os.path.join(self.download_folder, "crossref_metadata.json.tar.gz")

    def download(self):
        """Download release.

        :return: None.
        """

        logging.info(f"Downloading from url: {self.url}")

        # Set API token header
        connection = BaseHook.get_connection(AirflowConns.CROSSREF)
        api_token = connection.password
        header = {"Crossref-Plus-API-Token": f"Bearer {api_token}"}

        # Download release
        with requests.get(self.url, headers=header, stream=True) as response:
            # Check if authorisation with the api token was successful or not, raise error if not successful
            if response.status_code != 200:
                raise ConnectionError(f"Error downloading file {self.url}, status_code={response.status_code}")

            # Open file for saving
            with open(self.download_path, "wb") as file:
                response.raw.read = functools.partial(response.raw.read, decode_content=True)
                shutil.copyfileobj(response.raw, file)

        logging.info(f"Successfully download url to {self.download_path}")

    def extract(self):
        """Extract release. Decompress and unzip file to multiple json files.

        :return: None.
        """
        logging.info(f"extract_release: {self.download_path}")

        # Run command using GNUtar, bsdtar (on e.g. OS x) might give error: 'Error inclusion pattern: Failed to open
        # 'pigz -d'
        cmd = f'tar -xv -I "pigz -d" -f {self.download_path} -C {self.extract_folder}'
        p: Popen = subprocess.Popen(
            cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, executable="/bin/bash"
        )
        stdout, stderr = wait_for_process(p)
        logging.debug(stdout)
        success = p.returncode == 0 and "error" not in stderr.lower()

        if success:
            logging.info(f"extract_release success: {self.download_path}")
        else:
            logging.error(stdout)
            logging.error(stderr)
            raise AirflowException(f"extract_release error: {self.download_path}")

    def transform(self, max_workers: int):
        """Transform the Crossref Metadata release.
        Each extracted file is transformed. This is done in parallel using the ThreadPoolExecutor.

        :param max_workers: the number of processes to use when transforming files (one process per file).
        :return: whether the transformation was successful or not.
        """
        logging.info(f"Transform input folder: {self.extract_folder}, output folder: {self.transform_folder}")
        finished = 0
        # Transform each file in parallel
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = []

            # List files and sort so that they are processed in ascending order
            input_file_paths = natsorted(self.extract_files)

            # Create tasks for each file
            for input_file in input_file_paths:
                # The output file will be a json lines file, hence adding the 'l' to the file extension
                output_file = os.path.join(self.transform_folder, os.path.basename(input_file) + "l")
                future = executor.submit(transform_file, input_file, output_file)
                futures.append(future)

            # Wait for completed tasks
            for future in as_completed(futures):
                future.result()
                finished += 1
                if finished % 1000 == 0:
                    logging.info(f"Transformed {finished} files")


class CrossrefMetadataTelescope(SnapshotTelescope):
    """
    The Crossref Metadata Telescope

    Saved to the BigQuery table: <project_id>.crossref.crossref_metadataYYYYMMDD
    """

    DAG_ID = "crossref_metadata"
    DATASET_ID = "crossref"
    SCHEDULE_INTERVAL = "0 0 7 * *"
    TELESCOPE_URL = "https://api.crossref.org/snapshots/monthly/{year}/{month:02d}/all.json.tar.gz"

    def __init__(
        self,
        dag_id: str = DAG_ID,
        start_date: pendulum.DateTime = pendulum.datetime(2020, 6, 7),
        schedule_interval: str = SCHEDULE_INTERVAL,
        dataset_id: str = "crossref",
        schema_folder: str = default_schema_folder(),
        queue: str = "remote_queue",
        dataset_description: str = "The Crossref Metadata Plus dataset: "
        "https://www.crossref.org/services/metadata-retrieval/metadata-plus/",
        load_bigquery_table_kwargs: Dict = None,
        table_descriptions: Dict = None,
        airflow_vars: List = None,
        airflow_conns: List = None,
        max_active_runs: int = 1,
        max_processes: int = min(32, os.cpu_count() + 4),
    ):
        """The Crossref Metadata telescope

        :param dag_id: the id of the DAG.
        :param start_date: the start date of the DAG.
        :param schedule_interval: the schedule interval of the DAG.
        :param dataset_id: the BigQuery dataset id.
        :param schema_folder: the SQL schema path.
        :param queue: Crossref Metadata tasks run on the worker VM, indicated by the 'remote_queue'.
        :param dataset_description: description for the BigQuery dataset.
        :param load_bigquery_table_kwargs: the customisation parameters for loading data into a BigQuery table.
        :param table_descriptions: a dictionary with table ids and corresponding table descriptions.
        :param airflow_vars: list of airflow variable keys, for each variable it is checked if it exists in airflow.
        :param airflow_conns: list of airflow connection keys, for each connection it is checked if it exists in airflow
        :param max_active_runs: the maximum number of DAG runs that can be run at once.
        :param max_processes: the number of max workers used in ThreadPoolExecutor to transform files in parallel.
        """

        if table_descriptions is None:
            table_descriptions = {dag_id: "A single Crossref Metadata snapshot."}

        if airflow_vars is None:
            airflow_vars = [
                AirflowVars.DATA_PATH,
                AirflowVars.PROJECT_ID,
                AirflowVars.DATA_LOCATION,
                AirflowVars.DOWNLOAD_BUCKET,
                AirflowVars.TRANSFORM_BUCKET,
            ]
        if airflow_conns is None:
            airflow_conns = [AirflowConns.CROSSREF]
        super().__init__(
            dag_id,
            start_date,
            schedule_interval,
            dataset_id,
            schema_folder,
            queue=queue,
            load_bigquery_table_kwargs=load_bigquery_table_kwargs,
            dataset_description=dataset_description,
            table_descriptions=table_descriptions,
            airflow_vars=airflow_vars,
            airflow_conns=airflow_conns,
            max_active_runs=max_active_runs,
        )
        self.max_processes = max_processes

        self.add_setup_task(self.check_dependencies)
        self.add_setup_task(self.check_release_exists)
        self.add_task(self.download)
        self.add_task(self.upload_downloaded)
        self.add_task(self.extract)
        self.add_task(self.transform)
        self.add_task(self.upload_transformed)
        self.add_task(self.bq_load)
        self.add_task(self.cleanup)

    def make_release(self, **kwargs) -> List[CrossrefMetadataRelease]:
        """Make release instances. The release is passed as an argument to the function (TelescopeFunction) that is
        called in 'task_callable'.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are
        passed to this argument.
        :return: a list of CrossrefMetadataRelease instances.
        """

        release_date = kwargs["execution_date"]
        return [CrossrefMetadataRelease(self.dag_id, release_date)]

    def check_release_exists(self, **kwargs):
        """Check that the release for this month exists.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """
        # List all available releases
        logging.info(f"Listing available releases since start date ({self.start_date}):")
        for dt in pendulum.period(pendulum.instance(self.start_date), pendulum.today("UTC")).range("years"):
            response = requests.get(f"https://api.crossref.org/snapshots/monthly/{dt.year}")
            soup = BeautifulSoup(response.text)
            hrefs = soup.find_all("a", href=True)
            for href in hrefs:
                logging.info(href["href"])

        # Construct the release for the execution date and check if it exists.
        # The release release for a given execution_date is added on the 5th day of the following month.
        # E.g. the 2020-05 release is added to the website on 2020-06-05.
        execution_date = kwargs["execution_date"]

        url = CrossrefMetadataTelescope.TELESCOPE_URL.format(year=execution_date.year, month=execution_date.month)
        logging.info(f"Checking if available release exists for {execution_date.year}-{execution_date.month}")

        response = retry_session().head(url)
        if response.status_code == 302:
            logging.info(f"Snapshot exists at url: {url}, response code: {response.status_code}")
            return True
        elif response.reason == "Not Found":
            logging.info(
                f"Snapshot does not exist at url: {url}, response code: {response.status_code}, "
                f"reason: {response.reason}"
            )
            return False
        else:
            raise AirflowException(
                f"Could not get head of url: {url}, response code: {response.status_code}," f"reason: {response.reason}"
            )

    def download(self, releases: List[CrossrefMetadataRelease], **kwargs):
        """Task to download the CrossrefMetadataRelease release for a given month.

        :param releases: the list of CrossrefMetadataRelease instances.
        :return: None.
        """

        # Download each release
        for release in releases:
            release.download()

    def upload_downloaded(self, releases: List[CrossrefMetadataRelease], **kwargs):
        """Task to upload the downloaded CrossrefMetadataRelease release for a given month.

        :param releases: the list of CrossrefMetadataRelease instances.
        :return: None.
        """

        for release in releases:
            upload_files_from_list(release.download_files, release.download_bucket)

    def extract(self, releases: List[CrossrefMetadataRelease], **kwargs):
        """Task to extract the CrossrefMetadataRelease release for a given month.

        :param releases: the list of CrossrefMetadataRelease instances.
        :return: None.
        """

        for release in releases:
            release.extract()

    def transform(self, releases: List[CrossrefMetadataRelease], **kwargs):
        """Task to transform the CrossrefMetadataRelease release for a given month.

        :param releases: the list of CrossrefMetadataRelease instances.
        :return: None.
        """

        for release in releases:
            release.transform(max_workers=self.max_processes)

    def bq_load(self, releases: List[SnapshotRelease], **kwargs):
        """Task to load each transformed release to BigQuery.
        The table_id is set to the file name without the extension.

        :param releases: a list of releases.
        :return: None.
        """

        # Load each transformed release
        for release in releases:
            transform_blob = f"{blob_name(release.transform_folder)}/*"
            table_description = self.table_descriptions.get(self.dag_id, "")
            bq_load_shard(
                self.schema_folder,
                release.release_date,
                transform_blob,
                self.dataset_id,
                self.dag_id,
                self.source_format,
                prefix=self.schema_prefix,
                schema_version=self.schema_version,
                dataset_description=self.dataset_description,
                table_description=table_description,
                **self.load_bigquery_table_kwargs,
            )


def transform_file(input_file_path: str, output_file_path: str):
    """Transform a single crossref metadata json file.
    The json file is converted to a jsonl file and field names are transformed so they are accepted by BigQuery.

    :param input_file_path: the path of the file to transform.
    :param output_file_path: where to save the transformed file.
    :return: None.
    """

    cmd = (
        'mawk \'BEGIN {FS="\\":";RS=",\\"";OFS=FS;ORS=RS} {for (i=1; i<=NF;i++) if(i != NF) gsub("-", "_", $i)}1\''
        f" {input_file_path} | "
        'mawk \'!/^\}$|^\]$|,\\"$/{gsub("\[\[", "[");gsub("]]", "]");gsub(/,[ \\t]*$/,"");'
        'gsub("\\"timestamp\\":_", "\\"timestamp\\":");gsub("\\"date_parts\\":\[null]", "\\"date_parts\\":[]");'
        'gsub(/^\{\\"items\\":\[/,"");print}\' > '
        f"{output_file_path}"
    )
    p: Popen = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, executable="/bin/bash")
    stdout, stderr = wait_for_process(p)
    if p.returncode != 0:
        raise AirflowException(
            f"transform unsuccessful for file: {input_file_path}. " f"stdout: {stdout}, stderr: {stderr}"
        )



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Tuan Chien


import logging

import pendulum
from airflow import DAG
from airflow.operators.python import PythonOperator
from airflow.operators.subdag import SubDagOperator
from academic_observatory_workflows.workflows.scopus_telescope import ScopusTelescope
from observatory.platform.utils.airflow_utils import list_connections

default_args = {
    "owner": "airflow",
    "start_date": pendulum.datetime(2018, 1, 1),
}


def subdag_factory(parent_dag_id, connection, args):
    """Factory for making the ETL subdags.

    :param parent_dag_id: parent dag's id.
    :param connection: Airflow Connection object.
    :param args: default arguments to use for this subdag.
    :return: DAG object.
    """

    institution = str(connection)[ScopusTelescope.ID_STRING_OFFSET :]
    logging.info(f"Spawning ETL subdag for: {institution}")

    subdag = DAG(
        dag_id=f"{parent_dag_id}.{institution}",
        default_args=args,
        catchup=False,
        schedule_interval=ScopusTelescope.SCHEDULE_INTERVAL,
    )

    with subdag:
        # Check that dependencies exist before starting
        check_dependencies = PythonOperator(
            task_id=ScopusTelescope.TASK_ID_CHECK_DEPENDENCIES,
            python_callable=ScopusTelescope.check_dependencies,
            op_kwargs={"dag_start": "{{dag_run.start_date}}", "conn": connection, "institution": institution},
            queue=ScopusTelescope.QUEUE,
        )

        download = PythonOperator(
            task_id=ScopusTelescope.TASK_ID_DOWNLOAD,
            python_callable=ScopusTelescope.download,
            op_kwargs={"conn": connection},
            queue=ScopusTelescope.QUEUE,
            retries=ScopusTelescope.RETRIES,
        )

        # Upload gzipped response (json)
        upload_downloaded = PythonOperator(
            task_id=ScopusTelescope.TASK_ID_UPLOAD_DOWNLOADED,
            python_callable=ScopusTelescope.upload_downloaded,
            queue=ScopusTelescope.QUEUE,
            retries=ScopusTelescope.RETRIES,
        )

        # Transform into database schema format (jsonline output)
        transform_db_format = PythonOperator(
            task_id=ScopusTelescope.TASK_ID_TRANSFORM_DB_FORMAT,
            python_callable=ScopusTelescope.transform_db_format,
            queue=ScopusTelescope.QUEUE,
        )

        # # Upload the gzipped transformed jsonline entries to Google Cloud Storage
        upload_transformed = PythonOperator(
            task_id=ScopusTelescope.TASK_ID_UPLOAD_TRANSFORMED,
            python_callable=ScopusTelescope.upload_transformed,
            queue=ScopusTelescope.QUEUE,
            retries=ScopusTelescope.RETRIES,
        )

        # Load the transformed SCOPUS snapshot to BigQuery
        # Depends on past so that BigQuery load jobs are not all created at once
        bq_load = PythonOperator(
            task_id=ScopusTelescope.TASK_ID_BQ_LOAD,
            python_callable=ScopusTelescope.bq_load,
            queue=ScopusTelescope.QUEUE,
        )

        cleanup = PythonOperator(
            task_id=ScopusTelescope.TASK_ID_CLEANUP,
            python_callable=ScopusTelescope.cleanup,
            queue=ScopusTelescope.QUEUE,
        )

        (
            check_dependencies
            >> download
            >> upload_downloaded
            >> transform_db_format
            >> upload_transformed
            >> bq_load
            >> cleanup
        )

    return subdag


with DAG(
    dag_id=ScopusTelescope.DAG_ID,
    schedule_interval=ScopusTelescope.SCHEDULE_INTERVAL,
    catchup=False,
    default_args=default_args,
) as dag:
    # Only process if the SCOPUS API server is up.
    check_api_server = PythonOperator(
        task_id=ScopusTelescope.TASK_CHECK_API_SERVER,
        python_callable=ScopusTelescope.check_api_server,
        queue=ScopusTelescope.QUEUE,
        retries=ScopusTelescope.RETRIES,
    )

    # Spawn a subdag to handle ETL for each institution
    subdags = list()
    conns = list_connections(ScopusTelescope.DAG_ID)
    for conn in conns:
        subdag = SubDagOperator(
            task_id=str(conn)[ScopusTelescope.ID_STRING_OFFSET :],
            subdag=subdag_factory(ScopusTelescope.DAG_ID, conn, default_args),
            default_args=default_args,
            dag=dag,
        )

        # Task dependencies
        check_api_server >> subdag
        subdags.append(subdag)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs

import os
from datetime import timedelta
from unittest.mock import patch

import pendulum
import vcr
from airflow.exceptions import AirflowSkipException
from click.testing import CliRunner

from academic_observatory_workflows.config import test_fixtures_folder
from academic_observatory_workflows.workflows.crossref_events_telescope import (
    CrossrefEventsRelease,
    CrossrefEventsTelescope,
    parse_event_url,
    transform_batch,
)
from observatory.platform.utils.test_utils import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    module_file_path,
)
from observatory.platform.utils.url_utils import get_user_agent
from observatory.platform.utils.workflow_utils import blob_name


class TestCrossrefEventsTelescope(ObservatoryTestCase):
    """Tests for the Crossref Events telescope"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.
        :param args: arguments.
        :param kwargs: keyword arguments.
        """
        super(TestCrossrefEventsTelescope, self).__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")

        self.first_execution_date = pendulum.datetime(year=2018, month=5, day=14)
        self.first_cassette = test_fixtures_folder("crossref_events", "crossref_events1.yaml")

        self.second_execution_date = pendulum.datetime(year=2018, month=5, day=20)
        self.second_cassette = test_fixtures_folder("crossref_events", "crossref_events2.yaml")

        # additional tests setup
        self.start_date = pendulum.datetime(2021, 5, 6)
        self.end_date = pendulum.datetime(2021, 5, 12)
        self.release = CrossrefEventsRelease(
            CrossrefEventsTelescope.DAG_ID,
            self.start_date,
            self.end_date,
            False,
            "mailto",
            max_threads=21,
            max_processes=1,
        )

    def test_dag_structure(self):
        """Test that the Crossref Events DAG has the correct structure.
        :return: None
        """

        dag = CrossrefEventsTelescope().make_dag()
        self.assert_dag_structure(
            {
                "check_dependencies": ["get_release_info"],
                "get_release_info": ["download"],
                "download": ["upload_downloaded"],
                "upload_downloaded": ["transform"],
                "transform": ["upload_transformed"],
                "upload_transformed": ["bq_load_partition"],
                "bq_load_partition": ["bq_delete_old"],
                "bq_delete_old": ["bq_append_new"],
                "bq_append_new": ["cleanup"],
                "cleanup": [],
            },
            dag,
        )

    def test_dag_load(self):
        """Test that the Crossref Events DAG can be loaded from a DAG bag.
        :return: None
        """

        with ObservatoryEnvironment().create():
            dag_file = os.path.join(
                module_file_path("academic_observatory_workflows.dags"), "crossref_events_telescope.py"
            )
            self.assert_dag_load("crossref_events", dag_file)

    def test_telescope(self):
        """Test the Crossref Events telescope end to end.
        :return: None.
        """
        # Setup Observatory environment
        env = ObservatoryEnvironment(self.project_id, self.data_location)
        dataset_id = env.add_dataset()

        # Setup Telescope
        telescope = CrossrefEventsTelescope(dataset_id=dataset_id)
        telescope.max_threads = 1
        telescope.max_processes = 1
        dag = telescope.make_dag()

        # Create the Observatory environment and run tests
        with env.create():
            # first run
            with env.create_dag_run(dag, self.first_execution_date):
                # Test that all dependencies are specified: no error should be thrown
                env.run_task(telescope.check_dependencies.__name__)

                # Test list releases task with files available
                ti = env.run_task(telescope.get_release_info.__name__)
                start_date, end_date, first_release = ti.xcom_pull(
                    key=CrossrefEventsTelescope.RELEASE_INFO,
                    task_ids=telescope.get_release_info.__name__,
                    include_prior_dates=False,
                )

                start_date = pendulum.parse(start_date)
                end_date = pendulum.parse(end_date)

                self.assertEqual(start_date, dag.default_args["start_date"])
                self.assertEqual(end_date, pendulum.today("UTC") - timedelta(days=1))
                self.assertTrue(first_release)

                # use release info for other tasks
                release = CrossrefEventsRelease(
                    telescope.dag_id,
                    start_date,
                    end_date,
                    first_release,
                    telescope.mailto,
                    telescope.max_threads,
                    telescope.max_processes,
                )

                # Test download task
                with vcr.use_cassette(self.first_cassette):
                    env.run_task(telescope.download.__name__)
                self.assertEqual(6, len(release.download_files))
                for file in release.download_files:
                    if "2018-05-14" in file:
                        download_hash = "9a18d1002a5395de3cbcd9c61fb28c83"
                    else:
                        download_hash = "ad9cf98aab232eee7edf12375f016770"
                    self.assert_file_integrity(file, download_hash, "md5")

                # Test that files uploaded
                env.run_task(telescope.upload_downloaded.__name__)
                for file in release.download_files:
                    self.assert_blob_integrity(env.download_bucket, blob_name(file), file)

                # Test that files transformed
                env.run_task(telescope.transform.__name__)
                self.assertEqual(6, len(release.transform_files))
                for file in release.transform_files:
                    if "2018-05-14" in file:
                        transform_hash = "3e953d2424fe37739790bbc5c2410824"
                    else:
                        transform_hash = "d5e0a887656d1786a9e7c4dbdbf77ba1"
                    self.assert_file_integrity(file, transform_hash, "md5")

                # Test that transformed files uploaded
                env.run_task(telescope.upload_transformed.__name__)
                for file in release.transform_files:
                    self.assert_blob_integrity(env.transform_bucket, blob_name(file), file)

                # Test that load partition task is skipped for the first release
                ti = env.run_task(telescope.bq_load_partition.__name__)
                self.assertEqual(ti.state, "skipped")

                # Test delete old task is in success state, without doing anything
                ti = env.run_task(telescope.bq_delete_old.__name__)
                self.assertEqual(ti.state, "success")

                # Test append new creates table
                env.run_task(telescope.bq_append_new.__name__)
                main_table_id, partition_table_id = release.dag_id, f"{release.dag_id}_partitions"
                table_id = f"{self.project_id}.{telescope.dataset_id}.{main_table_id}"
                expected_rows = 68
                self.assert_table_integrity(table_id, expected_rows)

                # Test that all telescope data deleted
                download_folder, extract_folder, transform_folder = (
                    release.download_folder,
                    release.extract_folder,
                    release.transform_folder,
                )
                env.run_task(telescope.cleanup.__name__)
                self.assert_cleanup(download_folder, extract_folder, transform_folder)

            # second run
            with env.create_dag_run(dag, self.second_execution_date):
                # Test that all dependencies are specified: no error should be thrown
                env.run_task(telescope.check_dependencies.__name__)

                # Test list releases task with files available
                ti = env.run_task(telescope.get_release_info.__name__)
                start_date, end_date, first_release = ti.xcom_pull(
                    key=CrossrefEventsTelescope.RELEASE_INFO,
                    task_ids=telescope.get_release_info.__name__,
                    include_prior_dates=False,
                )
                start_date = pendulum.parse(start_date)
                end_date = pendulum.parse(end_date)
                self.assertEqual(release.end_date + timedelta(days=1), start_date)
                self.assertEqual(pendulum.today("UTC") - timedelta(days=1), end_date)
                self.assertFalse(first_release)

                # use release info for other tasks
                release = CrossrefEventsRelease(
                    telescope.dag_id,
                    start_date,
                    end_date,
                    first_release,
                    telescope.mailto,
                    telescope.max_threads,
                    telescope.max_processes,
                )

                # Test download task
                with vcr.use_cassette(self.second_cassette):
                    env.run_task(telescope.download.__name__)

                self.assertEqual(20, len(release.download_files))
                for file in release.download_files:
                    if "edited" in file:
                        download_hash = "b1c8c856c29365efeeef8a7c1ccba7da"
                    elif "deleted" in file:
                        download_hash = "8d52425faa9192e8748865b8c53c2b3d"
                    else:
                        download_hash = "01aa964587e6296df5697d13a122e8ce"
                    self.assert_file_integrity(file, download_hash, "md5")

                # Test that file uploaded
                env.run_task(telescope.upload_downloaded.__name__)
                for file in release.download_files:
                    self.assert_blob_integrity(env.download_bucket, blob_name(file), file)

                # Test that file transformed
                env.run_task(telescope.transform.__name__)

                self.assertEqual(20, len(release.transform_files))
                for file in release.transform_files:
                    if "edited" in file:
                        transform_hash = "902437a731a4aed529f4e0d176d2222b"
                    elif "deleted" in file:
                        transform_hash = "10b6d1911aaaad14204d867884722da4"
                    else:
                        transform_hash = "513d71d356d8356d1365d1dd25b1f71a"
                    self.assert_file_integrity(file, transform_hash, "md5")

                # Test that transformed file uploaded
                env.run_task(telescope.upload_transformed.__name__)
                for file in release.transform_files:
                    self.assert_blob_integrity(env.transform_bucket, blob_name(file), file)

                # Test that load partition task creates partition
                env.run_task(telescope.bq_load_partition.__name__)
                main_table_id, partition_table_id = release.dag_id, f"{release.dag_id}_partitions"
                table_id = f'{self.project_id}.{telescope.dataset_id}.{partition_table_id}${pendulum.today().strftime("%Y%m%d")}'
                expected_rows = 82
                self.assert_table_integrity(table_id, expected_rows)

                # Test task deleted rows from main table
                env.run_task(telescope.bq_delete_old.__name__)
                table_id = f"{self.project_id}.{telescope.dataset_id}.{main_table_id}"
                expected_rows = 60
                self.assert_table_integrity(table_id, expected_rows)

                # Test append new adds rows to table
                env.run_task(telescope.bq_append_new.__name__)
                table_id = f"{self.project_id}.{telescope.dataset_id}.{main_table_id}"
                expected_rows = 142
                self.assert_table_integrity(table_id, expected_rows)

                # Test that all telescope data deleted
                download_folder, extract_folder, transform_folder = (
                    release.download_folder,
                    release.extract_folder,
                    release.transform_folder,
                )
                env.run_task(telescope.cleanup.__name__)
                self.assert_cleanup(download_folder, extract_folder, transform_folder)

    def test_urls(self):
        """Test the urls property of release
        :return: None.
        """
        events_url = (
            "https://api.eventdata.crossref.org/v1/events?mailto={mail_to}"
            "&from-collected-date={start_date}&until-collected-date={end_date}&rows=1000"
        )
        edited_url = (
            "https://api.eventdata.crossref.org/v1/events/edited?"
            "mailto={mail_to}&from-updated-date={start_date}"
            "&until-updated-date={end_date}&rows=1000"
        )
        deleted_url = (
            "https://api.eventdata.crossref.org/v1/events/deleted?"
            "mailto={mail_to}&from-updated-date={start_date}"
            "&until-updated-date={end_date}&rows=1000"
        )

        self.release.first_release = True
        urls = self.release.urls
        self.assertEqual(7, len(urls))
        for url in urls:
            event_type, date = parse_event_url(url)
            self.assertEqual(event_type, "events")
            expected_url = events_url.format(mail_to=self.release.mailto, start_date=date, end_date=date)
            self.assertEqual(expected_url, url)

        self.release.first_release = False
        urls = self.release.urls
        self.assertEqual(21, len(urls))
        for url in urls:
            event_type, date = parse_event_url(url)
            if event_type == "events":
                expected_url = events_url.format(mail_to=self.release.mailto, start_date=date, end_date=date)
            elif event_type == "edited":
                expected_url = edited_url.format(mail_to=self.release.mailto, start_date=date, end_date=date)
            else:
                expected_url = deleted_url.format(mail_to=self.release.mailto, start_date=date, end_date=date)
            self.assertEqual(expected_url, url)

    @patch.object(CrossrefEventsRelease, "download_batch")
    @patch("observatory.platform.utils.workflow_utils.Variable.get")
    def test_download(self, mock_variable_get, mock_download_batch):
        """Test the download method of the release in parallel mode
        :return: None.
        """
        mock_variable_get.return_value = "data"
        with CliRunner().isolated_filesystem():
            # Test download without any events returned
            with self.assertRaises(AirflowSkipException):
                self.release.download()

            # Test download with events returned
            mock_download_batch.reset_mock()
            events_path = os.path.join(self.release.download_folder, "events.jsonl")
            with open(events_path, "w") as f:
                f.write("[{'test': 'test'}]\n")

            self.release.download()
            self.assertEqual(len(self.release.urls), mock_download_batch.call_count)

    @patch("academic_observatory_workflows.workflows.crossref_events_telescope.download_events")
    @patch("observatory.platform.utils.workflow_utils.Variable.get")
    def test_download_batch(self, mock_variable_get, mock_download_events):
        """Test download_batch function
        :return: None.
        """
        mock_variable_get.return_value = os.path.join(os.getcwd(), "data")
        self.release.first_release = True
        batch_number = 0
        url = self.release.urls[batch_number]
        headers = {"User-Agent": get_user_agent(package_name="academic_observatory_workflows")}
        with CliRunner().isolated_filesystem():
            events_path = self.release.batch_path(url)
            cursor_path = self.release.batch_path(url, cursor=True)

            # Test with existing cursor path
            with open(cursor_path, "w") as f:
                f.write("cursor")
            mock_download_events.return_value = (None, 10, 10)
            self.release.download_batch(batch_number, url)
            self.assertFalse(os.path.exists(cursor_path))
            mock_download_events.assert_called_once_with(url, headers, events_path, cursor_path)

            # Test with no existing previous files
            mock_download_events.reset_mock()
            mock_download_events.return_value = (None, 10, 10)
            self.release.download_batch(batch_number, url)
            mock_download_events.assert_called_once_with(url, headers, events_path, cursor_path)

            # Test with events path and no cursor path, so previous successful attempt
            mock_download_events.reset_mock()
            with open(events_path, "w") as f:
                f.write("events")
            self.release.download_batch(batch_number, url)
            mock_download_events.assert_not_called()
            os.remove(events_path)

    @patch("observatory.platform.utils.workflow_utils.Variable.get")
    def test_transform_batch(self, mock_variable_get):
        """Test the transform_batch method of the release
        :return: None.
        """

        with CliRunner().isolated_filesystem() as t:
            mock_variable_get.return_value = os.path.join(t, "data")

            # Use release info so that we can download the right data
            release = CrossrefEventsRelease(
                "crossref_events",
                pendulum.datetime(2018, 5, 14),
                pendulum.datetime(2018, 5, 19),
                True,
                "aniek.roelofs@curtin.edu.au",
                max_threads=1,
                max_processes=1,
            )

            # Download files
            with vcr.use_cassette(self.first_cassette):
                release.download()

            # Transform batch
            for file_path in release.download_files:
                transform_batch(file_path, release.transform_folder)

            # Assert all transformed
            self.assertEqual(len(release.download_files), len(release.transform_files))



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs

import os
from typing import List
from unittest.mock import patch

import httpretty
import pendulum
import vcr
from airflow.utils.state import State
from click.testing import CliRunner

from academic_observatory_workflows.config import test_fixtures_folder
from academic_observatory_workflows.workflows.crossref_fundref_telescope import (
    CrossrefFundrefRelease,
    CrossrefFundrefTelescope,
    list_releases,
    strip_whitespace,
)
from observatory.platform.utils.file_utils import _hash_file
from observatory.platform.utils.test_utils import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    module_file_path,
)
from observatory.platform.utils.workflow_utils import (
    bigquery_sharded_table_id,
    blob_name,
)


class TestCrossrefFundrefTelescope(ObservatoryTestCase):
    """Tests for the CrossrefFundref telescope"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """

        super(TestCrossrefFundrefTelescope, self).__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")
        self.download_path = test_fixtures_folder("crossref_fundref", "crossref_fundref_v1.34.tar.gz")
        self.download_hash = "0cd65042"
        self.extract_hash = "559aa89d41a85ff84d705084c1caeb8d"
        self.transform_hash = "632b453a"

    def test_dag_structure(self):
        """Test that the CrossrefFundref DAG has the correct structure.

        :return: None
        """
        # mock create_pool to prevent querying non existing airflow db
        with patch("academic_observatory_workflows.workflows.crossref_fundref_telescope.create_pool"):
            dag = CrossrefFundrefTelescope().make_dag()
            self.assert_dag_structure(
                {
                    "check_dependencies": ["get_release_info"],
                    "get_release_info": ["download"],
                    "download": ["upload_downloaded"],
                    "upload_downloaded": ["extract"],
                    "extract": ["transform"],
                    "transform": ["upload_transformed"],
                    "upload_transformed": ["bq_load"],
                    "bq_load": ["cleanup"],
                    "cleanup": [],
                },
                dag,
            )

    def test_dag_load(self):
        """Test that the CrossrefFundref DAG can be loaded from a DAG bag.

        :return: None
        """

        with ObservatoryEnvironment().create():
            dag_file = os.path.join(module_file_path("academic_observatory_workflows.dags"), "crossref_fundref_telescope.py")
            self.assert_dag_load("crossref_fundref", dag_file)

    def test_telescope(self):
        """Test the CrossrefFundref telescope end to end.

        :return: None.
        """

        # Setup Observatory environment
        env = ObservatoryEnvironment(self.project_id, self.data_location)
        dataset_id = env.add_dataset()

        # Create the Observatory environment and run tests
        with env.create():
            # Setup Telescope inside env, so pool can be created
            execution_date = pendulum.datetime(year=2021, month=6, day=1)
            telescope = CrossrefFundrefTelescope(dataset_id=dataset_id)
            dag = telescope.make_dag()

            with env.create_dag_run(dag, execution_date):
                # Test that all dependencies are specified: no error should be thrown
                ti = env.run_task(telescope.check_dependencies.__name__, dag, execution_date)
                self.assertEqual(ti.state, State.SUCCESS)

                # Test list releases task
                release_info = [
                    {
                        "url": "https://gitlab.com/crossref/open_funder_registry/-/archive/v1.34/open_funder_registry-v1.34.tar.gz",
                        "date": pendulum.parse("2021-05-19T09:34:09.898000+00:00"),
                    }
                ]
                with patch("academic_observatory_workflows.workflows.crossref_fundref_telescope.list_releases") as mock_list_releases:
                    mock_list_releases.return_value = release_info
                    ti = env.run_task(telescope.get_release_info.__name__, dag, execution_date)

                actual_release_info = ti.xcom_pull(
                    key=CrossrefFundrefTelescope.RELEASE_INFO,
                    task_ids=telescope.get_release_info.__name__,
                    include_prior_dates=False,
                )
                self.assertEqual(release_info, actual_release_info)

                # Create release instance to check results from other tasks
                release = CrossrefFundrefRelease(
                    telescope.dag_id, pendulum.parse(release_info[0]["date"]), release_info[0]["url"]
                )

                # Test download task
                with httpretty.enabled():
                    self.setup_mock_file_download(release.url, self.download_path)
                    env.run_task(telescope.download.__name__, dag, execution_date)
                self.assertEqual(1, len(release.download_files))
                self.assert_file_integrity(release.download_path, self.download_hash, "gzip_crc")

                # Test that file uploaded
                env.run_task(telescope.upload_downloaded.__name__, dag, execution_date)
                self.assert_blob_integrity(env.download_bucket, blob_name(release.download_path), release.download_path)

                # Test that file extracted
                env.run_task(telescope.extract.__name__, dag, execution_date)
                self.assertEqual(1, len(release.extract_files))
                self.assert_file_integrity(release.extract_path, self.extract_hash, "md5")

                # Test that file transformed
                env.run_task(telescope.transform.__name__, dag, execution_date)
                self.assertEqual(1, len(release.transform_files))
                self.assert_file_integrity(release.transform_path, self.transform_hash, "gzip_crc")

                # Test that transformed file uploaded
                env.run_task(telescope.upload_transformed.__name__, dag, execution_date)
                self.assert_blob_integrity(
                    env.transform_bucket, blob_name(release.transform_path), release.transform_path
                )

                # Test that data loaded into BigQuery
                env.run_task(telescope.bq_load.__name__, dag, execution_date)
                table_id = (
                    f"{self.project_id}.{dataset_id}."
                    f"{bigquery_sharded_table_id(telescope.dag_id, release.release_date)}"
                )
                expected_rows = 27949
                self.assert_table_integrity(table_id, expected_rows)

                # Test that all telescope data deleted
                download_folder, extract_folder, transform_folder = (
                    release.download_folder,
                    release.extract_folder,
                    release.transform_folder,
                )
                env.run_task(telescope.cleanup.__name__, dag, execution_date)
                self.assert_cleanup(download_folder, extract_folder, transform_folder)

    def test_list_releases(self):
        """Test that list releases returns a list with dictionaries of release info.

        :return: None.
        """
        cassette_path = test_fixtures_folder("crossref_fundref", "list_fundref_releases.yaml")
        with vcr.use_cassette(cassette_path):
            releases = list_releases(pendulum.datetime(2014, 3, 1), pendulum.datetime(2020, 6, 1))
            self.assertIsInstance(releases, List)
            self.assertEqual(39, len(releases))
            for release in releases:
                self.assertIsInstance(release, dict)
                self.assertIsInstance(release["url"], str)
                self.assertIsInstance(pendulum.parse(release["date"]), pendulum.DateTime)

    def test_strip_whitespace(self):
        with CliRunner().isolated_filesystem():
            # Create file with space
            file_with_space = "file1.txt"
            with open(file_with_space, "w") as f:
                f.write(" ")
                f.write("test")

            # Create file without space and store hash
            file_without_space = "file2.txt"
            with open(file_without_space, "w") as f:
                f.write("test")
            expected_hash = _hash_file(file_without_space, algorithm="md5")

            # Strip whitespace and check that files are now the same
            strip_whitespace(file_with_space)
            self.assert_file_integrity(file_with_space, expected_hash, "md5")

            # Check that file stays the same when first line is not a space
            strip_whitespace(file_without_space)
            self.assert_file_integrity(file_without_space, expected_hash, "md5")



# Copyright 2020, 2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

# The keywords airflow and DAG are required to load the DAGs from this file, see bullet 2 in the Apache Airflow FAQ:
# https://airflow.apache.org/docs/stable/faq.html

import json
import os
from typing import Callable, Dict, List

from academic_observatory_workflows.config import elastic_mappings_folder
from observatory.platform.elastic.elastic import KeepInfo, KeepOrder
from observatory.platform.elastic.kibana import TimeField
from observatory.platform.utils.jinja2_utils import render_template
from observatory.platform.utils.workflow_utils import make_dag_id
from observatory.platform.workflows.elastic_import_workflow import (
    ElasticImportConfig,
    ElasticImportWorkflow,
    load_elastic_mappings_simple,
)

DATASET_ID = "data_export"
DATA_LOCATION = "us"
FILE_TYPE_JSONL = "jsonl.gz"
DAG_ONIX_WORKFLOW_PREFIX = "onix_workflow"
DAG_PREFIX = "elastic_import"
ELASTIC_MAPPINGS_PATH = elastic_mappings_folder()
AO_KIBANA_TIME_FIELDS = [TimeField("^.*$", "published_year")]

# These can be customised per DAG.  Just using some generic settings for now.
index_keep_info = {
    "": KeepInfo(ordering=KeepOrder.newest, num=2),
    "ao": KeepInfo(ordering=KeepOrder.newest, num=2),
}


def load_elastic_mappings_ao(path: str, table_prefix: str, simple_prefixes: List = None):
    """For the Observatory project, load the Elastic mappings for a given table_prefix.
    :param path: the path to the mappings files.
    :param table_prefix: the table_id prefix (without shard date).
    :param simple_prefixes: the prefixes of mappings to load with the load_elastic_mappings_simple function.
    :return: the rendered mapping as a Dict.
    """

    # Set default simple_prefixes
    if simple_prefixes is None:
        simple_prefixes = ["ao_doi"]

    if not table_prefix.startswith("ao"):
        raise ValueError("Table must begin with 'ao'")
    elif any([table_prefix.startswith(prefix) for prefix in simple_prefixes]):
        return load_elastic_mappings_simple(path, table_prefix)
    else:
        prefix, aggregate, facet = table_prefix.split("_", 2)
        mappings_file_name = "ao-relations-mappings.json.jinja2"
        is_fixed_facet = facet in ["unique_list", "access_types", "disciplines", "output_types", "events", "metrics"]
        if is_fixed_facet:
            mappings_file_name = f"ao-{facet.replace('_', '-')}-mappings.json.jinja2"
        mappings_path = os.path.join(path, mappings_file_name)
        return json.loads(render_template(mappings_path, aggregate=aggregate, facet=facet))


configs = [
    ElasticImportConfig(
        dag_id=make_dag_id(DAG_PREFIX, "observatory"),
        project_id="academic-observatory",
        dataset_id=DATASET_ID,
        bucket_name="academic-observatory-transform",
        data_location=DATA_LOCATION,
        file_type=FILE_TYPE_JSONL,
        sensor_dag_ids=["doi"],
        kibana_spaces=["coki-scratch-space", "coki-dashboards", "dev-coki-dashboards"],
        elastic_mappings_path=ELASTIC_MAPPINGS_PATH,
        elastic_mappings_func=load_elastic_mappings_ao,
        kibana_time_fields=AO_KIBANA_TIME_FIELDS,
        index_keep_info=index_keep_info,
    )
]

for config in configs:
    dag = ElasticImportWorkflow(
        dag_id=config.dag_id,
        project_id=config.project_id,
        dataset_id=config.dataset_id,
        bucket_name=config.bucket_name,
        data_location=config.data_location,
        file_type=config.file_type,
        sensor_dag_ids=config.sensor_dag_ids,
        elastic_mappings_folder=ELASTIC_MAPPINGS_PATH,
        elastic_mappings_func=config.elastic_mappings_func,
        kibana_spaces=config.kibana_spaces,
        kibana_time_fields=config.kibana_time_fields,
        index_keep_info=config.index_keep_info,
    ).make_dag()
    globals()[dag.dag_id] = dag



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

"""
A DAG that harvests the OpenCitations Indexes: http://opencitations.net/

Saved to the BigQuery table: <project_id>.open_citations.open_citationsYYYYMMDD
"""

import pendulum
from airflow import DAG
from airflow.operators.python import PythonOperator, ShortCircuitOperator
from academic_observatory_workflows.workflows.open_citations_telescope import OpenCitationsTelescope

default_args = {"owner": "airflow", "start_date": pendulum.datetime(2018, 7, 1)}

with DAG(dag_id=OpenCitationsTelescope.DAG_ID, schedule_interval="@weekly", default_args=default_args) as dag:
    # Check that dependencies exist before starting
    check_task = PythonOperator(
        task_id=OpenCitationsTelescope.TASK_ID_CHECK_DEPENDENCIES,
        python_callable=OpenCitationsTelescope.check_dependencies,
        queue=OpenCitationsTelescope.QUEUE,
    )

    # List releases and skip all subsequent tasks if there is no release to process
    list_releases_task = ShortCircuitOperator(
        task_id=OpenCitationsTelescope.TASK_ID_LIST_RELEASES,
        python_callable=OpenCitationsTelescope.list_releases,
        queue=OpenCitationsTelescope.QUEUE,
    )

    # Download the Open Citations releases for a given interval
    download_task = PythonOperator(
        task_id=OpenCitationsTelescope.TASK_ID_DOWNLOAD,
        python_callable=OpenCitationsTelescope.download,
        queue=OpenCitationsTelescope.QUEUE,
        retries=OpenCitationsTelescope.RETRIES,
    )

    # Upload the Open Citations releases for a given interval
    upload_downloaded = PythonOperator(
        task_id=OpenCitationsTelescope.TASK_ID_UPLOAD_DOWNLOADED,
        python_callable=OpenCitationsTelescope.upload_downloaded,
        queue=OpenCitationsTelescope.QUEUE,
        retries=OpenCitationsTelescope.RETRIES,
    )

    # Extract the Open Citations releases for a given interval
    extract_task = PythonOperator(
        task_id=OpenCitationsTelescope.TASK_ID_EXTRACT,
        python_callable=OpenCitationsTelescope.extract,
        queue=OpenCitationsTelescope.QUEUE,
    )

    # Upload the transformed Open Citations releases for a given interval to Google Cloud Storage
    upload_extracted_task = PythonOperator(
        task_id=OpenCitationsTelescope.TASK_ID_UPLOAD_EXTRACTED,
        python_callable=OpenCitationsTelescope.upload_extracted,
        queue=OpenCitationsTelescope.QUEUE,
        retries=OpenCitationsTelescope.RETRIES,
    )

    # Load the transformed Open Citations releases for a given interval to BigQuery
    bq_load_task = PythonOperator(
        task_id=OpenCitationsTelescope.TASK_ID_BQ_LOAD,
        python_callable=OpenCitationsTelescope.load_to_bq,
        queue=OpenCitationsTelescope.QUEUE,
    )

    # Delete all files
    cleanup_task = PythonOperator(
        task_id=OpenCitationsTelescope.TASK_ID_CLEANUP,
        python_callable=OpenCitationsTelescope.cleanup,
        queue=OpenCitationsTelescope.QUEUE,
    )

    # Task dependencies
    (
        check_task
        >> list_releases_task
        >> download_task
        >> upload_downloaded
        >> extract_task
        >> upload_extracted_task
        >> bq_load_task
        >> cleanup_task
    )



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import glob
import json
import logging
import os
import shutil
import subprocess
from dataclasses import dataclass
from typing import List

import pendulum
from airflow.exceptions import AirflowException
from airflow.models.taskinstance import TaskInstance
from airflow.models.variable import Variable
from google.cloud.bigquery import SourceFormat

from academic_observatory_workflows.config import schema_folder
from observatory.platform.utils.airflow_utils import AirflowVars, check_variables
from observatory.platform.utils.config_utils import find_schema
from observatory.platform.utils.data_utils import get_file
from observatory.platform.utils.gc_utils import (
    bigquery_sharded_table_id,
    bigquery_table_exists,
    create_bigquery_dataset,
    load_bigquery_table,
    upload_files_to_cloud_storage,
)
from observatory.platform.utils.proc_utils import wait_for_process
from observatory.platform.utils.url_utils import retry_session
from observatory.platform.utils.workflow_utils import SubFolder, workflow_path

OPEN_CITATIONS_ARTICLE_ID = 6741422
OPEN_CITATIONS_VERSION_URL = "https://api.figshare.com/v2/articles/{article_id}/versions"


@dataclass
class File:
    name: str
    download_url: str
    md5hash: str
    parent: "OpenCitationsRelease" = None

    @property
    def download_blob_name(self):
        return f"telescopes/{OpenCitationsTelescope.DAG_ID}/{self.parent.release_name}/{self.name}"


class OpenCitationsRelease:
    release_date: pendulum.DateTime
    files: List[File]

    def __init__(self, release_date: pendulum.DateTime, files: List[File]):
        self.release_date = release_date
        self.files = files

        # Add parent release to files
        for file in self.files:
            file.parent = self

    @property
    def release_name(self) -> str:
        return f'{OpenCitationsTelescope.DAG_ID}_{self.release_date.strftime("%Y_%m_%d")}'

    @property
    def download_path(self) -> str:
        return os.path.join(workflow_path(SubFolder.downloaded, OpenCitationsTelescope.DAG_ID), self.release_name)

    @property
    def extract_path(self) -> str:
        return os.path.join(workflow_path(SubFolder.extracted, OpenCitationsTelescope.DAG_ID), self.release_name)

    @property
    def transformed_blob_path(self) -> str:
        return f"telescopes/{OpenCitationsTelescope.DAG_ID}/{self.release_name}/*.csv"


def list_open_citations_releases(
    start_date: pendulum.DateTime = None, end_date: pendulum.DateTime = None, timeout: float = 30.0
) -> List[OpenCitationsRelease]:
    """List the Open Citations releases for a given time period.

    :param start_date: the start date.
    :param end_date: the end date.
    :param timeout: timeout in seconds.
    :return: a list of Open Citations releases found within the given time period (if the time period was specified).
    """

    versions = fetch_open_citations_versions()
    releases = []
    for version in versions:
        version_url = version["url"]
        response = retry_session().get(version_url, timeout=timeout, headers={"Accept-encoding": "gzip"})
        article = json.loads(response.text)
        release_date = pendulum.parse(article["created_date"])

        if (start_date is None or start_date <= release_date) and (end_date is None or release_date <= end_date):
            files = []
            for file in article["files"]:
                name = file["name"]
                download_url = file["download_url"]
                md5hash = file["computed_md5"]
                files.append(File(name, download_url, md5hash))
            releases.append(OpenCitationsRelease(release_date, files))
    return releases


def fetch_open_citations_versions():
    """Fetch the Open Citation versions that are available.

    :return: the open citations versions.
    """

    url = OPEN_CITATIONS_VERSION_URL.format(article_id=OPEN_CITATIONS_ARTICLE_ID)
    response = retry_session().get(url, timeout=30, headers={"Accept-encoding": "gzip"})
    return json.loads(response.text)


def pull_releases(ti: TaskInstance) -> List[OpenCitationsRelease]:
    return ti.xcom_pull(
        key=OpenCitationsTelescope.RELEASES_TOPIC_NAME,
        task_ids=OpenCitationsTelescope.TASK_ID_LIST_RELEASES,
        include_prior_dates=False,
    )


class OpenCitationsTelescope:
    """A container for holding the constants and static functions for the Open Citations telescope."""

    DAG_ID = "open_citations"
    DESCRIPTION = "The OpenCitations Indexes: http://opencitations.net/"
    DATASET_ID = DAG_ID
    QUEUE = "remote_queue"
    RETRIES = 3
    RELEASES_TOPIC_NAME = "releases"

    TASK_ID_CHECK_DEPENDENCIES = "check_dependencies"
    TASK_ID_LIST_RELEASES = f"list_releases"
    TASK_ID_DOWNLOAD = f"download"
    TASK_ID_UPLOAD_DOWNLOADED = f"upload_downloaded"
    TASK_ID_EXTRACT = f"extract"
    TASK_ID_UPLOAD_EXTRACTED = f"upload_extracted"
    TASK_ID_BQ_LOAD = f"bq_load"
    TASK_ID_CLEANUP = f"cleanup"

    @staticmethod
    def check_dependencies(**kwargs):
        """Check that all variables exist that are required to run the DAG.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        """

        vars_valid = check_variables(
            AirflowVars.DATA_PATH,
            AirflowVars.PROJECT_ID,
            AirflowVars.DATA_LOCATION,
            AirflowVars.DOWNLOAD_BUCKET,
            AirflowVars.TRANSFORM_BUCKET,
        )
        if not vars_valid:
            raise AirflowException("Required variables are missing")

    @staticmethod
    def list_releases(**kwargs):
        """Task to lists all Open Citations releases for a given time period.

        :param kwargs: the context passed from the BranchPythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: the identifier of the task to execute next.
        """

        start_date = kwargs["execution_date"]
        end_date = kwargs["next_execution_date"].subtract(microseconds=1)
        releases = list_open_citations_releases(start_date=start_date, end_date=end_date)
        project_id = Variable.get(AirflowVars.PROJECT_ID)

        # Check if we can skip any releases
        releases_out = []
        for release in releases:
            table_id = bigquery_sharded_table_id(OpenCitationsTelescope.DAG_ID, release.release_date)

            if bigquery_table_exists(project_id, OpenCitationsTelescope.DATASET_ID, table_id):
                logging.info(
                    f"Skipping as table exists for {release.release_name} release: "
                    f"{project_id}.{OpenCitationsTelescope.DATASET_ID}.{table_id}"
                )
            else:
                logging.info(
                    f"Table doesn't exist yet, processing Open Citations {release.release_date} "
                    f"release in this workflow"
                )
                releases_out.append(release)

        continue_dag = len(releases_out)
        if continue_dag:
            # Push messages
            ti: TaskInstance = kwargs["ti"]
            ti.xcom_push(OpenCitationsTelescope.RELEASES_TOPIC_NAME, releases_out, start_date)
        return continue_dag

    @staticmethod
    def download(**kwargs):
        """Task to download the Open Citations releases for a given time period.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Pull messages
        ti: TaskInstance = kwargs["ti"]
        releases = pull_releases(ti)

        # Download and extract each release posted this month
        for release in releases:
            os.makedirs(release.download_path, exist_ok=True)
            for file in release.files:
                file_path, updated = get_file(
                    file.name,
                    file.download_url,
                    cache_subdir="",
                    cache_dir=release.download_path,
                    md5_hash=file.md5hash,
                )

    @staticmethod
    def upload_downloaded(**kwargs):
        """Task to upload the downloaded Open Citations releases for a given month.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Get bucket name
        bucket_name = Variable.get(AirflowVars.DOWNLOAD_BUCKET)

        # Pull messages
        ti: TaskInstance = kwargs["ti"]
        releases = pull_releases(ti)

        # Upload each release
        file_paths = []
        blob_names = []
        for release in releases:
            for file in release.files:
                file_path = os.path.join(release.download_path, file.name)
                blob_name = file.download_blob_name
                file_paths.append(file_path)
                blob_names.append(blob_name)

        success = upload_files_to_cloud_storage(bucket_name, blob_names, file_paths)
        if not success:
            raise AirflowException("Problem uploading files")

    @staticmethod
    def extract(**kwargs):
        """Task to extract the downloaded Open Citations releases for a given month.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Pull messages
        ti: TaskInstance = kwargs["ti"]
        releases = pull_releases(ti)

        for release in releases:
            os.makedirs(release.extract_path, exist_ok=True)

            for file in release.files:
                file_path = os.path.join(release.download_path, file.name)
                cmd = f"unzip {file_path} -d {release.extract_path}"
                logging.info(f"Running command: {cmd}")
                p = subprocess.Popen(
                    cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, executable="/bin/bash"
                )
                stdout, stderr = wait_for_process(p)

                if stdout:
                    logging.info(stdout)

                if stderr:
                    raise AirflowException(f"bash command failed for {file_path}, {release.release_date}: {stderr}")

                logging.info(f"File extracted to: {release.extract_path}")

    @staticmethod
    def upload_extracted(**kwargs):
        """Task to upload the extracted Open Citations releases for a given month.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Get bucket name
        bucket_name = Variable.get(AirflowVars.TRANSFORM_BUCKET)

        # Pull messages
        ti: TaskInstance = kwargs["ti"]
        releases = pull_releases(ti)

        # Upload each release
        file_paths = []
        blob_names = []
        for release in releases:
            release_file_paths = glob.glob(f"{release.extract_path}/*.csv")
            for file_path in release_file_paths:
                file_name = os.path.basename(file_path)
                blob_name = f"telescopes/{OpenCitationsTelescope.DAG_ID}/{release.release_name}/{file_name}"
                file_paths.append(file_path)
                blob_names.append(blob_name)

        logging.info(f"file_paths: {file_paths}")
        logging.info(f"blob_names: {blob_names}")

        success = upload_files_to_cloud_storage(bucket_name, blob_names, file_paths)
        if not success:
            raise AirflowException("Problem uploading files")

    @staticmethod
    def load_to_bq(**kwargs):
        """Load transformed Open Citations release into BigQuery.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Pull messages
        ti: TaskInstance = kwargs["ti"]
        releases = pull_releases(ti)

        # Get variables
        project_id = Variable.get(AirflowVars.PROJECT_ID)
        data_location = Variable.get(AirflowVars.DATA_LOCATION)
        bucket_name = Variable.get(AirflowVars.TRANSFORM_BUCKET)

        # Create dataset
        dataset_id = OpenCitationsTelescope.DATASET_ID
        create_bigquery_dataset(project_id, dataset_id, data_location, OpenCitationsTelescope.DESCRIPTION)
        table_name = OpenCitationsTelescope.DAG_ID

        # Load each release
        for release in releases:
            table_id = bigquery_sharded_table_id(table_name, release.release_date)

            # Select schema file based on release date
            analysis_schema_path = schema_folder()
            schema_file_path = find_schema(analysis_schema_path, table_name, release.release_date)
            if schema_file_path is None:
                logging.error(
                    f"No schema found with search parameters: analysis_schema_path={analysis_schema_path}, "
                    f"table_name={table_name}, release_date={release.release_date}"
                )
                exit(os.EX_CONFIG)

            # Load BigQuery table
            uri = f"gs://{bucket_name}/{release.transformed_blob_path}"
            logging.info(f"URI: {uri}")
            success = load_bigquery_table(
                uri,
                dataset_id,
                data_location,
                table_id,
                schema_file_path,
                SourceFormat.CSV,
                csv_field_delimiter=",",
                csv_quote_character='"',
                csv_skip_leading_rows=1,
                csv_allow_quoted_newlines=True,
            )
            if not success:
                raise AirflowException("bq_load task: data failed to load data into BigQuery")

    @staticmethod
    def cleanup(**kwargs):
        """Delete files of downloaded and extracted Open Citations releases.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Pull messages
        ti: TaskInstance = kwargs["ti"]
        releases = pull_releases(ti)

        for release in releases:
            logging.info(f"Removing downloaded files for release: {release.release_name}")
            download_path = release.download_path
            try:
                shutil.rmtree(download_path)
            except FileNotFoundError as e:
                logging.warning(f"No such file or directory {download_path}: {e}")

            logging.info(f"Removing extracted files for release: {release.release_name}")
            extract_path = release.extract_path
            try:
                shutil.rmtree(extract_path)
            except FileNotFoundError as e:
                logging.warning(f"No such file or directory {extract_path}: {e}")



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs


import os
from datetime import datetime
from unittest.mock import patch

import httpretty
import pendulum
from airflow.exceptions import AirflowException
from airflow.models.connection import Connection
from natsort import natsorted

from academic_observatory_workflows.config import test_fixtures_folder
from academic_observatory_workflows.workflows.crossref_metadata_telescope import (
    CrossrefMetadataRelease,
    CrossrefMetadataTelescope,
    transform_file,
)
from observatory.platform.utils.airflow_utils import AirflowConns
from observatory.platform.utils.gc_utils import bigquery_sharded_table_id
from observatory.platform.utils.test_utils import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    module_file_path,
)
from observatory.platform.utils.workflow_utils import blob_name


class TestCrossrefMetadataTelescope(ObservatoryTestCase):
    """Tests for the Crossref Metadata telescope"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """

        super(TestCrossrefMetadataTelescope, self).__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")
        self.download_path = test_fixtures_folder("crossref_metadata", "crossref_metadata.json.tar.gz")
        self.extract_file_hashes = [
            "4a55065d90aaa58c69bc5f5a54da3006",
            "c45901a52154789470410aad51485e9c",
            "4c0fd617224a557b9ef04313cca0bd4a",
            "d93dc613e299871925532d906c3a44a1",
            "dd1ab247c55191a14bcd1bf32719c337",
        ]
        self.transform_hashes = [
            "065a8c0bd1ef4239be9f37c0ad199a31",
            "38b766ec494054e621787de00ff715c8",
            "70437aad7c4568ed07408baf034871e4",
            "c3e3285a48867c8b7c10b1c9c0c5ab8a",
            "71ba3612352bcb2a723d4aa33ec35b61",
        ]

        # release used for tests outside observatory test environment
        self.release = CrossrefMetadataRelease("crossref_metadata", datetime(2020, 1, 1))

    def test_dag_structure(self):
        """Test that the Crossref Metadata DAG has the correct structure.

        :return: None
        """

        dag = CrossrefMetadataTelescope().make_dag()
        self.assert_dag_structure(
            {
                "check_dependencies": ["check_release_exists"],
                "check_release_exists": ["download"],
                "download": ["upload_downloaded"],
                "upload_downloaded": ["extract"],
                "extract": ["transform"],
                "transform": ["upload_transformed"],
                "upload_transformed": ["bq_load"],
                "bq_load": ["cleanup"],
                "cleanup": [],
            },
            dag,
        )

    def test_dag_load(self):
        """Test that the Crossref Metadata DAG can be loaded from a DAG bag.

        :return: None
        """

        with ObservatoryEnvironment().create():
            dag_file = os.path.join(module_file_path("academic_observatory_workflows.dags"), "crossref_metadata_telescope.py")
            self.assert_dag_load("crossref_metadata", dag_file)

    def test_telescope(self):
        """Test the Crossref Metadata telescope end to end.

        :return: None.
        """

        # Setup Observatory environment
        env = ObservatoryEnvironment(self.project_id, self.data_location)
        dataset_id = env.add_dataset()

        # Setup Telescope
        execution_date = pendulum.datetime(year=2020, month=11, day=1)
        telescope = CrossrefMetadataTelescope(dataset_id=dataset_id)
        dag = telescope.make_dag()

        # Create the Observatory environment and run tests
        with env.create():
            with env.create_dag_run(dag, execution_date):
                # Add Crossref Metadata connection
                env.add_connection(Connection(conn_id=AirflowConns.CROSSREF, uri="mysql://:crossref-token@"))

                # Test that all dependencies are specified: no error should be thrown
                env.run_task(telescope.check_dependencies.__name__, dag, execution_date)

                # Test check release exists task, next tasks should not be skipped
                with httpretty.enabled():
                    url = CrossrefMetadataTelescope.TELESCOPE_URL.format(
                        year=execution_date.year, month=execution_date.month
                    )
                    httpretty.register_uri(httpretty.HEAD, url, body="", status=302)
                    env.run_task(telescope.check_release_exists.__name__, dag, execution_date)

                release = CrossrefMetadataRelease(telescope.dag_id, execution_date)

                # Test download task
                with httpretty.enabled():
                    self.setup_mock_file_download(release.url, self.download_path)
                    env.run_task(telescope.download.__name__, dag, execution_date)
                self.assertEqual(1, len(release.download_files))
                expected_file_hash = "10210c33936f9ba6b7e053f6f457591b"
                self.assert_file_integrity(release.download_path, expected_file_hash, "md5")

                # Test that file uploaded
                env.run_task(telescope.upload_downloaded.__name__, dag, execution_date)
                self.assert_blob_integrity(env.download_bucket, blob_name(release.download_path), release.download_path)

                # Test that file extracted
                env.run_task(telescope.extract.__name__, dag, execution_date)
                self.assertEqual(5, len(release.extract_files))
                for i, file in enumerate(natsorted(release.extract_files)):
                    expected_file_hash = self.extract_file_hashes[i]
                    self.assert_file_integrity(file, expected_file_hash, "md5")

                # Test that files transformed
                env.run_task(telescope.transform.__name__, dag, execution_date)
                self.assertEqual(5, len(release.transform_files))
                for i, file in enumerate(natsorted(release.transform_files)):
                    expected_file_hash = self.transform_hashes[i]
                    self.assert_file_integrity(file, expected_file_hash, "md5")

                # Test that transformed files uploaded
                env.run_task(telescope.upload_transformed.__name__, dag, execution_date)
                for file in release.transform_files:
                    self.assert_blob_integrity(env.transform_bucket, blob_name(file), file)

                # Test that data loaded into BigQuery
                env.run_task(telescope.bq_load.__name__, dag, execution_date)
                table_id = (
                    f"{self.project_id}.{dataset_id}."
                    f"{bigquery_sharded_table_id(telescope.dag_id, release.release_date)}"
                )
                expected_rows = 20
                self.assert_table_integrity(table_id, expected_rows)

                # Test that all telescope data deleted
                download_folder, extract_folder, transform_folder = (
                    release.download_folder,
                    release.extract_folder,
                    release.transform_folder,
                )
                env.run_task(telescope.cleanup.__name__, dag, execution_date)
                self.assert_cleanup(download_folder, extract_folder, transform_folder)

    @patch("academic_observatory_workflows.workflows.crossref_metadata_telescope.BaseHook.get_connection")
    def test_download(self, mock_conn):
        """Test download method of release with failing response

        :param mock_conn: Mock Airflow crossref connection
        :return: None.
        """
        mock_conn.return_value = Connection(AirflowConns.CROSSREF, "http://:crossref-token@")
        release = self.release
        with httpretty.enabled():
            httpretty.register_uri(httpretty.GET, release.url, body="", status=400)
            with self.assertRaises(ConnectionError):
                release.download()

    @patch("academic_observatory_workflows.workflows.crossref_metadata_telescope.subprocess.Popen")
    @patch("observatory.platform.utils.workflow_utils.Variable.get")
    def test_extract(self, mock_variable_get, mock_subprocess):
        """Test extract method of release with failing extract command

        :param mock_variable_get: Mock Airflow data path variable
        :param mock_subprocess: Mock the subprocess output
        :return: None.
        """
        mock_variable_get.return_value = "data"
        release = self.release

        mock_subprocess().returncode = 1
        mock_subprocess().communicate.return_value = "stdout".encode(), "stderr".encode()
        with self.assertRaises(AirflowException):
            release.extract()

    def test_check_release_exists(self):
        """Test the 'check_release_exists' task with different responses.

        :return: None.
        """
        release = self.release
        telescope = CrossrefMetadataTelescope()
        with httpretty.enabled():
            # register 3 responses, successful, release not found and 'other'
            httpretty.register_uri(
                httpretty.HEAD,
                uri=release.url,
                responses=[
                    httpretty.Response(body="", status=302),
                    httpretty.Response(body="", status=404, adding_headers={"reason": "Not Found"}),
                    httpretty.Response(body="", status=400),
                ],
            )

            continue_dag = telescope.check_release_exists(execution_date=release.release_date)
            self.assertTrue(continue_dag)

            continue_dag = telescope.check_release_exists(execution_date=release.release_date)
            self.assertFalse(continue_dag)

            with self.assertRaises(AirflowException):
                telescope.check_release_exists(execution_date=release.release_date)

    @patch("academic_observatory_workflows.workflows.crossref_metadata_telescope.subprocess.Popen")
    def test_transform_file(self, mock_subprocess):
        """Test transform_file function with failing transform command.

        :param mock_subprocess: Mock the subprocess output
        :return: None.
        """
        mock_subprocess().returncode = 1
        mock_subprocess().communicate.return_value = "stdout".encode(), "stderr".encode()
        with self.assertRaises(AirflowException):
            transform_file("input_file_path", "output_file_path")



# Copyright 2020-2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Richard Hosking, James Diprose

from __future__ import annotations

import logging
import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from datetime import timedelta
from typing import Dict, List, Optional, Tuple

import pendulum
from academic_observatory_workflows.config import sql_folder
from airflow.exceptions import AirflowException
from airflow.models import Variable
from observatory.platform.utils.airflow_utils import AirflowVars, set_task_state
from observatory.platform.utils.dag_run_sensor import DagRunSensor
from observatory.platform.utils.gc_utils import (
    bigquery_sharded_table_id,
    copy_bigquery_table,
    create_bigquery_dataset,
    create_bigquery_table_from_query,
    create_bigquery_view,
    select_table_shard_dates,
)
from observatory.platform.utils.jinja2_utils import (
    make_sql_jinja2_filename,
    render_template,
)
from observatory.platform.workflows.workflow import Workflow

MAX_QUERIES = 100


@dataclass
class Table:
    dataset_id: str
    table_id: str = None
    sharded: bool = False
    release_date: pendulum.DateTime = None


@dataclass
class Transform:
    inputs: Dict = None
    output_table: Table = None
    output_cluster: bool = False
    output_clustering_fields: List = None


@dataclass
class Aggregation:
    table_id: str
    aggregation_field: str
    group_by_time_field: str = "published_year"
    relate_to_institutions: bool = False
    relate_to_countries: bool = False
    relate_to_groups: bool = False
    relate_to_members: bool = False
    relate_to_journals: bool = False
    relate_to_funders: bool = False
    relate_to_publishers: bool = False


def make_dataset_transforms(
    dataset_id_crossref_events: str = "crossref",
    dataset_id_crossref_metadata: str = "crossref",
    dataset_id_crossref_fundref: str = "crossref",
    dataset_id_grid: str = "digital_science",
    dataset_id_iso: str = "iso",
    dataset_id_mag: str = "mag",
    dataset_id_orcid: str = "orcid",
    dataset_id_open_citations: str = "open_citations",
    dataset_id_unpaywall: str = "our_research",
    dataset_id_settings: str = "settings",
    dataset_id_observatory: str = "observatory",
    dataset_id_observatory_intermediate: str = "observatory_intermediate",
) -> Tuple[List[Transform], Transform, Transform]:
    return (
        [
            Transform(
                inputs={"crossref_events": Table(dataset_id_crossref_events, "crossref_events")},
                output_table=Table(dataset_id_observatory_intermediate, "crossref_events"),
                output_cluster=True,
                output_clustering_fields=["doi"],
            ),
            Transform(
                inputs={
                    "crossref_fundref": Table(dataset_id_crossref_fundref, "crossref_fundref", sharded=True),
                    "crossref_metadata": Table(dataset_id_crossref_metadata, "crossref_metadata", sharded=True),
                },
                output_table=Table(dataset_id_observatory_intermediate, "crossref_fundref"),
                output_cluster=True,
                output_clustering_fields=["doi"],
            ),
            Transform(
                inputs={
                    "grid": Table(dataset_id_grid, "grid", sharded=True),
                    "iso": Table(dataset_id_iso),
                    "settings": Table(dataset_id_settings),
                },
                output_table=Table(dataset_id_observatory_intermediate, "grid"),
            ),
            Transform(
                inputs={
                    "mag": Table(dataset_id_mag, "Affiliations", sharded=True),
                    "settings": Table(dataset_id_settings),
                },
                output_table=Table(dataset_id_observatory_intermediate, "mag"),
                output_cluster=True,
                output_clustering_fields=["Doi"],
            ),
            Transform(
                inputs={"orcid": Table(dataset_id_orcid, "orcid")},
                output_table=Table(dataset_id_observatory_intermediate, "orcid"),
                output_cluster=True,
                output_clustering_fields=["doi"],
            ),
            Transform(
                inputs={"open_citations": Table(dataset_id_open_citations, "open_citations", sharded=True)},
                output_table=Table(dataset_id_observatory_intermediate, "open_citations"),
                output_cluster=True,
                output_clustering_fields=["doi"],
            ),
            Transform(
                inputs={"unpaywall": Table(dataset_id_unpaywall, "unpaywall", sharded=True)},
                output_table=Table(dataset_id_observatory_intermediate, "unpaywall"),
                output_cluster=True,
                output_clustering_fields=["doi"],
            ),
        ],
        Transform(
            inputs={
                "observatory_intermediate": Table(dataset_id_observatory_intermediate),
                "unpaywall": Table(dataset_id_unpaywall),
                "crossref_metadata": Table(dataset_id_crossref_metadata, "crossref_metadata", sharded=True),
                "settings": Table(dataset_id_settings),
            },
            output_table=Table(dataset_id_observatory, "doi"),
            output_cluster=True,
            output_clustering_fields=["doi"],
        ),
        Transform(
            inputs={
                "observatory": Table(dataset_id_observatory, "doi", sharded=True),
                "crossref_events": Table(dataset_id_observatory_intermediate, "crossref_events", sharded=True),
            },
            output_table=Table(dataset_id_observatory, "book"),
            output_cluster=True,
            output_clustering_fields=["isbn"],
        ),
    )


def make_elastic_tables(
    aggregate_table_id: str,
    relate_to_institutions: bool = False,
    relate_to_countries: bool = False,
    relate_to_groups: bool = False,
    relate_to_members: bool = False,
    relate_to_journals: bool = False,
    relate_to_funders: bool = False,
    relate_to_publishers: bool = False,
):
    # Always export
    tables = [
        {
            "file_name": DoiWorkflow.EXPORT_UNIQUE_LIST_FILENAME,
            "aggregate": aggregate_table_id,
            "facet": "unique_list",
        },
        {
            "file_name": DoiWorkflow.EXPORT_ACCESS_TYPES_FILENAME,
            "aggregate": aggregate_table_id,
            "facet": "access_types",
        },
        {
            "file_name": DoiWorkflow.EXPORT_DISCIPLINES_FILENAME,
            "aggregate": aggregate_table_id,
            "facet": "disciplines",
        },
        {
            "file_name": DoiWorkflow.EXPORT_OUTPUT_TYPES_FILENAME,
            "aggregate": aggregate_table_id,
            "facet": "output_types",
        },
        {"file_name": DoiWorkflow.EXPORT_EVENTS_FILENAME, "aggregate": aggregate_table_id, "facet": "events"},
        {"file_name": DoiWorkflow.EXPORT_METRICS_FILENAME, "aggregate": aggregate_table_id, "facet": "metrics"},
    ]

    # Optional Relationships
    if relate_to_institutions:
        tables.append(
            {
                "file_name": DoiWorkflow.EXPORT_RELATIONS_FILENAME,
                "aggregate": aggregate_table_id,
                "facet": "institutions",
            }
        )
    if relate_to_countries:
        tables.append(
            {
                "file_name": DoiWorkflow.EXPORT_RELATIONS_FILENAME,
                "aggregate": aggregate_table_id,
                "facet": "countries",
            }
        )
    if relate_to_groups:
        tables.append(
            {
                "file_name": DoiWorkflow.EXPORT_RELATIONS_FILENAME,
                "aggregate": aggregate_table_id,
                "facet": "groupings",
            }
        )
    if relate_to_members:
        tables.append(
            {
                "file_name": DoiWorkflow.EXPORT_RELATIONS_FILENAME,
                "aggregate": aggregate_table_id,
                "facet": "members",
            }
        )
    if relate_to_journals:
        tables.append(
            {
                "file_name": DoiWorkflow.EXPORT_RELATIONS_FILENAME,
                "aggregate": aggregate_table_id,
                "facet": "journals",
            }
        )

    if relate_to_funders:
        tables.append(
            {
                "file_name": DoiWorkflow.EXPORT_RELATIONS_FILENAME,
                "aggregate": aggregate_table_id,
                "facet": "funders",
            }
        )

    if relate_to_publishers:
        tables.append(
            {
                "file_name": DoiWorkflow.EXPORT_RELATIONS_FILENAME,
                "aggregate": aggregate_table_id,
                "facet": "publishers",
            }
        )

    return tables


class DoiWorkflow(Workflow):
    INT_DATASET_ID = "observatory_intermediate"
    INT_DATASET_DESCRIPTION = "Intermediate processing dataset for the Academic Observatory."
    DASHBOARDS_DATASET_ID = "coki_dashboards"
    DASHBOARDS_DATASET_DESCRIPTION = "The latest data for display in the COKI dashboards."
    FINAL_DATASET_ID = "observatory"
    FINAL_DATASET_DESCRIPTION = "The Academic Observatory dataset."
    ELASTIC_DATASET_ID = "data_export"
    ELASTIC_DATASET_ID_DATASET_DESCRIPTION = "The Academic Observatory dataset for Elasticsearch."

    AGGREGATE_DOI_FILENAME = make_sql_jinja2_filename("aggregate_doi")
    EXPORT_UNIQUE_LIST_FILENAME = make_sql_jinja2_filename("export_unique_list")
    EXPORT_ACCESS_TYPES_FILENAME = make_sql_jinja2_filename("export_access_types")
    EXPORT_DISCIPLINES_FILENAME = make_sql_jinja2_filename("export_disciplines")
    EXPORT_EVENTS_FILENAME = make_sql_jinja2_filename("export_events")
    EXPORT_METRICS_FILENAME = make_sql_jinja2_filename("export_metrics")
    EXPORT_OUTPUT_TYPES_FILENAME = make_sql_jinja2_filename("export_output_types")
    EXPORT_RELATIONS_FILENAME = make_sql_jinja2_filename("export_relations")

    SENSOR_DAG_IDS = [
        "crossref_metadata",
        "crossref_fundref",
        "geonames",
        "grid",
        "mag",
        "open_citations",
        "unpaywall",
        "orcid",
        "crossref_events",
    ]

    AGGREGATIONS = [
        Aggregation(
            "country",
            "countries",
            relate_to_members=True,
            relate_to_journals=True,
            relate_to_funders=True,
            relate_to_publishers=True,
        ),
        Aggregation(
            "funder",
            "funders",
            relate_to_institutions=True,
            relate_to_countries=True,
            relate_to_groups=True,
            relate_to_members=True,
            relate_to_funders=True,
            relate_to_publishers=True,
        ),
        Aggregation(
            "group",
            "groupings",
            relate_to_institutions=True,
            relate_to_members=True,
            relate_to_journals=True,
            relate_to_funders=True,
            relate_to_publishers=True,
        ),
        Aggregation(
            "institution",
            "institutions",
            relate_to_institutions=True,
            relate_to_countries=True,
            relate_to_journals=True,
            relate_to_funders=True,
            relate_to_publishers=True,
        ),
        Aggregation(
            "author",
            "authors",
            relate_to_institutions=True,
            relate_to_countries=True,
            relate_to_groups=True,
            relate_to_journals=True,
            relate_to_funders=True,
            relate_to_publishers=True,
        ),
        Aggregation(
            "journal",
            "journals",
            relate_to_institutions=True,
            relate_to_countries=True,
            relate_to_groups=True,
            relate_to_journals=True,
            relate_to_funders=True,
        ),
        Aggregation(
            "publisher",
            "publishers",
            relate_to_institutions=True,
            relate_to_countries=True,
            relate_to_groups=True,
            relate_to_funders=True,
        ),
        Aggregation("region", "regions", relate_to_funders=True, relate_to_publishers=True),
        Aggregation("subregion", "subregions", relate_to_funders=True, relate_to_publishers=True),
    ]

    def __init__(
        self,
        *,
        intermediate_dataset_id: str = INT_DATASET_ID,
        dashboards_dataset_id: str = DASHBOARDS_DATASET_ID,
        observatory_dataset_id: str = FINAL_DATASET_ID,
        elastic_dataset_id: str = ELASTIC_DATASET_ID,
        transforms: Tuple = None,
        dag_id: Optional[str] = "doi",
        start_date: Optional[pendulum.DateTime] = pendulum.datetime(2020, 8, 30),
        schedule_interval: Optional[str] = "@weekly",
        catchup: Optional[bool] = False,
        airflow_vars: List = None,
    ):
        """Create the DoiWorkflow.
        :param intermediate_dataset_id: the BigQuery intermediate dataset id.
        :param dashboards_dataset_id: the BigQuery dashboards dataset id.
        :param observatory_dataset_id: the BigQuery observatory dataset id.
        :param elastic_dataset_id: the BigQuery elastic dataset id.
        :param dag_id: the DAG id.
        :param start_date: the start date.
        :param schedule_interval: the schedule interval.
        :param catchup: whether to catchup.
        :param airflow_vars: the required Airflow Variables.
        """

        if airflow_vars is None:
            airflow_vars = [
                AirflowVars.DATA_PATH,
                AirflowVars.PROJECT_ID,
                AirflowVars.DATA_LOCATION,
                AirflowVars.DOWNLOAD_BUCKET,
                AirflowVars.TRANSFORM_BUCKET,
            ]

        # Initialise Telesecope base class
        super().__init__(
            dag_id=dag_id,
            start_date=start_date,
            schedule_interval=schedule_interval,
            catchup=catchup,
            airflow_vars=airflow_vars,
        )

        self.intermediate_dataset_id = intermediate_dataset_id
        self.dashboards_dataset_id = dashboards_dataset_id
        self.observatory_dataset_id = observatory_dataset_id
        self.elastic_dataset_id = elastic_dataset_id

        if transforms is None:
            self.transforms, self.transform_doi, self.transform_book = make_dataset_transforms(
                dataset_id_observatory=observatory_dataset_id
            )
        else:
            self.transforms, self.transform_doi, self.transform_book = transforms

        self.create_tasks()

    def create_tasks(self):
        # Add sensors
        for ext_dag_id in self.SENSOR_DAG_IDS:
            sensor = DagRunSensor(
                task_id=f"{ext_dag_id}_sensor",
                external_dag_id=ext_dag_id,
                mode="reschedule",
                duration=timedelta(days=7),  # Look back up to 7 days from execution date
                poke_interval=int(timedelta(hours=1).total_seconds()),  # Check at this interval if dag run is ready
                timeout=int(timedelta(days=2).total_seconds()),  # Sensor will fail after 2 days of waiting
            )
            self.add_sensor(sensor)

        # Setup tasks
        self.add_setup_task(self.check_dependencies)

        # Create datasets
        self.add_task(self.create_datasets)

        # Create tasks for processing intermediate tables
        with self.parallel_tasks():
            for transform in self.transforms:
                task_id = f"create_{transform.output_table.table_id}"
                self.add_task(self.create_intermediate_table, op_kwargs={"transform": transform}, task_id=task_id)

        # Create DOI Table
        self.add_task(
            self.create_intermediate_table,
            op_kwargs={"transform": self.transform_doi},
            task_id=f"create_{self.transform_doi.output_table.table_id}",
        )

        # Create Book Table
        self.add_task(
            self.create_intermediate_table,
            op_kwargs={"transform": self.transform_book},
            task_id=f"create_{self.transform_book.output_table.table_id}",
        )

        # Create final tables
        with self.parallel_tasks():
            for agg in self.AGGREGATIONS:
                task_id = f"create_{agg.table_id}"
                self.add_task(
                    self.create_aggregate_table, op_kwargs={"aggregation": agg, "task_id": task_id}, task_id=task_id
                )

        # Copy tables and create views
        self.add_task(self.copy_to_dashboards)
        self.add_task(self.create_dashboard_views)

        # Export for Elastic
        with self.parallel_tasks():
            for agg in self.AGGREGATIONS:
                task_id = f"export_{agg.table_id}"
                self.add_task(
                    self.export_for_elastic, op_kwargs={"aggregation": agg, "task_id": task_id}, task_id=task_id
                )

    def make_release(self, **kwargs) -> ObservatoryRelease:
        """Make a release instance. The release is passed as an argument to the function (TelescopeFunction) that is
        called in 'task_callable'.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: A release instance or list of release instances
        """

        release_date = kwargs["next_execution_date"].subtract(microseconds=1).date()
        project_id = Variable.get(AirflowVars.PROJECT_ID)
        data_location = Variable.get(AirflowVars.DATA_LOCATION)

        return ObservatoryRelease(
            project_id=project_id,
            data_location=data_location,
            release_date=release_date,
            intermediate_dataset_id=self.intermediate_dataset_id,
            observatory_dataset_id=self.observatory_dataset_id,
            dashboards_dataset_id=self.dashboards_dataset_id,
            elastic_dataset_id=self.elastic_dataset_id,
        )

    def create_datasets(self, release: ObservatoryRelease, **kwargs):
        """Create required BigQuery datasets.

        :param release: the ObservatoryRelease.
        :param kwargs: the context passed from the Airflow Operator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: None.
        """

        release.create_datasets()

    def create_intermediate_table(self, release: ObservatoryRelease, **kwargs):
        """Create an intermediate table.

        :param release: the ObservatoryRelease.
        :param kwargs: the context passed from the Airflow Operator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: None.
        """

        transform: Transform = kwargs["transform"]
        release.create_intermediate_table(
            inputs=transform.inputs,
            output_dataset_id=transform.output_table.dataset_id,
            output_table_id=transform.output_table.table_id,
            output_cluster=transform.output_cluster,
            output_clustering_fields=transform.output_clustering_fields,
        )

    def create_aggregate_table(self, release: ObservatoryRelease, **kwargs):
        """Runs the aggregate table query.

        :param release: the ObservatoryRelease.
        :param kwargs: the context passed from the Airflow Operator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: None.
        """

        agg: Aggregation = kwargs["aggregation"]
        success = release.create_aggregate_table(
            aggregation_field=agg.aggregation_field,
            table_id=agg.table_id,
            group_by_time_field=agg.group_by_time_field,
            relate_to_institutions=agg.relate_to_institutions,
            relate_to_countries=agg.relate_to_countries,
            relate_to_groups=agg.relate_to_groups,
            relate_to_members=agg.relate_to_members,
            relate_to_journals=agg.relate_to_journals,
            relate_to_funders=agg.relate_to_funders,
            relate_to_publishers=agg.relate_to_publishers,
        )
        set_task_state(success, kwargs["task_id"])

    def copy_to_dashboards(self, release: ObservatoryRelease, **kwargs):
        """Copy tables to dashboards dataset.

        :param release: the ObservatoryRelease.
        :param kwargs: the context passed from the Airflow Operator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: None.
        """

        success = release.copy_to_dashboards()
        set_task_state(success, self.copy_to_dashboards.__name__)

    def create_dashboard_views(self, release: ObservatoryRelease, **kwargs):
        """Create views for dashboards dataset.

        :param release: the ObservatoryRelease.
        :param kwargs: the context passed from the Airflow Operator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: None.
        """

        release.create_dashboard_views()

    def export_for_elastic(self, release: ObservatoryRelease, **kwargs):
        """Export data in a de-nested form for Elasticsearch.

        :param release: the ObservatoryRelease.
        :param kwargs: the context passed from the Airflow Operator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: None.
        """

        agg = kwargs["aggregation"]
        success = release.export_for_elastic(
            table_id=agg.table_id,
            relate_to_institutions=agg.relate_to_institutions,
            relate_to_countries=agg.relate_to_countries,
            relate_to_groups=agg.relate_to_groups,
            relate_to_members=agg.relate_to_members,
            relate_to_journals=agg.relate_to_journals,
            relate_to_funders=agg.relate_to_funders,
            relate_to_publishers=agg.relate_to_publishers,
        )
        set_task_state(success, kwargs["task_id"])


class ObservatoryRelease:
    def __init__(
        self,
        *,
        project_id: str,
        data_location: str,
        release_date: pendulum.DateTime,
        intermediate_dataset_id: str,
        dashboards_dataset_id: str,
        observatory_dataset_id: str,
        elastic_dataset_id: str,
    ):
        """Construct an ObservatoryRelease.

        :param project_id: the Google Cloud project id.
        :param data_location: the location for BigQuery datasets.
        :param release_date: the release date.
        :param intermediate_dataset_id: the BigQuery intermediate dataset id.
        :param dashboards_dataset_id: the BigQuery dashboards dataset id.
        :param observatory_dataset_id: the BigQuery observatory dataset id.
        :param elastic_dataset_id: the BigQuery elastic dataset id.
        """

        self.project_id = project_id
        self.data_location = data_location
        self.release_date = release_date
        self.intermediate_dataset_id = intermediate_dataset_id
        self.dashboards_dataset_id = dashboards_dataset_id
        self.observatory_dataset_id = observatory_dataset_id
        self.elastic_dataset_id = elastic_dataset_id

    def create_datasets(self):
        """Create the BigQuery datasets where data will be saved.
        :return: None.
        """

        datasets = [
            (self.intermediate_dataset_id, DoiWorkflow.INT_DATASET_DESCRIPTION),
            (self.dashboards_dataset_id, DoiWorkflow.DASHBOARDS_DATASET_DESCRIPTION),
            (self.observatory_dataset_id, DoiWorkflow.FINAL_DATASET_DESCRIPTION),
            (self.elastic_dataset_id, DoiWorkflow.ELASTIC_DATASET_ID_DATASET_DESCRIPTION),
        ]

        for dataset_id, description in datasets:
            create_bigquery_dataset(
                self.project_id,
                dataset_id,
                self.data_location,
                description=description,
            )

    def create_intermediate_table(
        self,
        *,
        inputs: Dict,
        output_dataset_id: str,
        output_table_id: str,
        output_cluster: bool,
        output_clustering_fields: List,
    ):
        """Create an intermediate table.
        :param inputs: the input datasets.
        :param output_dataset_id: the output dataset id.
        :param output_table_id: the output table id.
        :param output_cluster: whether to cluster or not.
        :param output_clustering_fields: the fields to cluster on.
        :return: None.
        """

        def get_release_date(dataset_id: str, table_id: str):
            # Get last table shard date before current end date
            table_shard_dates = select_table_shard_dates(self.project_id, dataset_id, table_id, self.release_date)
            if len(table_shard_dates):
                shard_date = table_shard_dates[0]
            else:
                raise AirflowException(
                    f"{self.project_id}.{dataset_id}.{table_id} "
                    f"with a table shard date <= {self.release_date} not found"
                )

            return shard_date

        for k, table in inputs.items():
            if table.sharded:
                table.release_date = get_release_date(table.dataset_id, table.table_id)

        # Create processed table
        template_path = os.path.join(sql_folder(), make_sql_jinja2_filename(f"create_{output_table_id}"))
        sql = render_template(template_path, project_id=self.project_id, release_date=self.release_date, **inputs)

        output_table_id_sharded = bigquery_sharded_table_id(output_table_id, self.release_date)
        success = create_bigquery_table_from_query(
            sql=sql,
            project_id=self.project_id,
            dataset_id=output_dataset_id,
            table_id=output_table_id_sharded,
            location=self.data_location,
            cluster=output_cluster,
            clustering_fields=output_clustering_fields,
        )

        return success

    def create_aggregate_table(
        self,
        *,
        aggregation_field: str,
        table_id: str,
        group_by_time_field: str = "published_year",
        relate_to_institutions: bool = False,
        relate_to_countries: bool = False,
        relate_to_groups: bool = False,
        relate_to_members: bool = False,
        relate_to_journals: bool = False,
        relate_to_funders: bool = False,
        relate_to_publishers: bool = False,
    ) -> bool:
        """Runs the aggregate table query.
        :param aggregation_field: the field to aggregate on, e.g. institution, publisher etc.
        :param group_by_time_field: either published_year or published_year_month depending on the granularity required for
        the time dimension
        :param table_id: the table id.
        :param relate_to_institutions: whether to generate the institution relationship output for this query
        :param relate_to_countries: whether to generate the countries relationship output for this query
        :param relate_to_groups: whether to generate the groups relationship output for this query
        :param relate_to_members: whether to generate the members relationship output for this query
        :param relate_to_journals: whether to generate the journals relationship output for this query
        :param relate_to_funders: whether to generate the funders relationship output for this query
        :param relate_to_publishers: whether to generate the publish relationship output for this query
        :return: None.
        """

        template_path = os.path.join(sql_folder(), make_sql_jinja2_filename("create_aggregate"))
        sql = render_template(
            template_path,
            project_id=self.project_id,
            dataset_id=self.observatory_dataset_id,
            release_date=self.release_date,
            aggregation_field=aggregation_field,
            group_by_time_field=group_by_time_field,
            relate_to_institutions=relate_to_institutions,
            relate_to_countries=relate_to_countries,
            relate_to_groups=relate_to_groups,
            relate_to_members=relate_to_members,
            relate_to_journals=relate_to_journals,
            relate_to_funders=relate_to_funders,
            relate_to_publishers=relate_to_publishers,
        )

        sharded_table_id = bigquery_sharded_table_id(table_id, self.release_date)
        success = create_bigquery_table_from_query(
            sql=sql,
            project_id=self.project_id,
            dataset_id=self.observatory_dataset_id,
            table_id=sharded_table_id,
            location=self.data_location,
            cluster=True,
            clustering_fields=["id"],
        )

        return success

    def copy_to_dashboards(self) -> bool:
        """Copy all tables in the observatory dataset to the dashboards dataset.
        :return: whether successful or not.
        """

        results = []
        table_ids = [agg.table_id for agg in DoiWorkflow.AGGREGATIONS] + ["doi"]
        for table_id in table_ids:
            source_table_id = f"{self.project_id}.{self.observatory_dataset_id}.{bigquery_sharded_table_id(table_id, self.release_date)}"
            destination_table_id = f"{self.project_id}.{self.dashboards_dataset_id}.{table_id}"
            success = copy_bigquery_table(source_table_id, destination_table_id, self.data_location)
            if not success:
                logging.error(f"Issue copying table: {source_table_id} to {destination_table_id}")

            results.append(success)

        return all(results)

    def create_dashboard_views(self):
        """Create views.
        :return: None.
        """

        # Create processed dataset
        template_path = os.path.join(sql_folder(), make_sql_jinja2_filename("comparison_view"))

        # Create views
        table_ids = ["country", "funder", "group", "institution", "publisher", "subregion"]
        for table_id in table_ids:
            view_name = f"{table_id}_comparison"
            query = render_template(
                template_path, project_id=self.project_id, dataset_id=self.dashboards_dataset_id, table_id=table_id
            )
            create_bigquery_view(self.project_id, self.dashboards_dataset_id, view_name, query)

    def export_for_elastic(
        self,
        *,
        table_id: str,
        relate_to_institutions: bool = False,
        relate_to_countries: bool = False,
        relate_to_groups: bool = False,
        relate_to_members: bool = False,
        relate_to_journals: bool = False,
        relate_to_funders: bool = False,
        relate_to_publishers: bool = False,
    ) -> bool:
        """Export data in in a de-nested form for elastic
        :param table_id:
        :param relate_to_institutions:
        :param relate_to_countries:
        :param relate_to_groups:
        :param relate_to_members:
        :param relate_to_journals:
        :param relate_to_funders:
        :param relate_to_publishers:
        :return: whether successful or not.
        """

        tables = make_elastic_tables(
            table_id,
            relate_to_institutions=relate_to_institutions,
            relate_to_countries=relate_to_countries,
            relate_to_groups=relate_to_groups,
            relate_to_members=relate_to_members,
            relate_to_journals=relate_to_journals,
            relate_to_funders=relate_to_funders,
            relate_to_publishers=relate_to_publishers,
        )

        # Calculate the number of parallel queries. Since all of the real work is done on BigQuery run each export task
        # in a separate thread so that they can be done in parallel.
        num_queries = min(len(tables), MAX_QUERIES)
        results = []

        with ThreadPoolExecutor(max_workers=num_queries) as executor:
            futures = list()
            futures_msgs = {}
            for table in tables:
                template_file_name = table["file_name"]
                aggregate = table["aggregate"]
                facet = table["facet"]

                msg = f"Exporting file_name={template_file_name}, aggregate={aggregate}, facet={facet}"
                logging.info(msg)
                future = executor.submit(
                    self.export_aggregate_table,
                    table_id=table_id,
                    template_file_name=template_file_name,
                    aggregate=aggregate,
                    facet=facet,
                )

                futures.append(future)
                futures_msgs[future] = msg

            # Wait for completed tasks
            for future in as_completed(futures):
                success = future.result()
                msg = futures_msgs[future]
                results.append(success)
                if success:
                    logging.info(f"Exporting feed success: {msg}")
                else:
                    logging.error(f"Exporting feed failed: {msg}")

        return all(results)

    def export_aggregate_table(self, *, table_id: str, template_file_name: str, aggregate: str, facet: str):
        """Export an aggregate table.
        :param table_id:
        :param template_file_name:
        :param aggregate:
        :param facet:
        :return:
        """

        template_path = os.path.join(sql_folder(), template_file_name)
        sql = render_template(
            template_path,
            project_id=self.project_id,
            dataset_id=self.observatory_dataset_id,
            table_id=table_id,
            release_date=self.release_date,
            aggregate=aggregate,
            facet=facet,
        )

        export_table_id = f"ao_{aggregate}_{facet}"
        processed_table_id = bigquery_sharded_table_id(export_table_id, self.release_date)

        success = create_bigquery_table_from_query(
            sql=sql,
            project_id=self.project_id,
            dataset_id=self.elastic_dataset_id,
            table_id=processed_table_id,
            location=self.data_location,
        )

        return success
